"""
Session management for Janito Chat CLI.
Defines ChatSession and ChatShellState classes.
"""

import types
from rich.console import Console
from rich.rule import Rule
from prompt_toolkit.history import InMemoryHistory
from janito.cli.chat_mode.shell.input_history import UserInputHistory
from prompt_toolkit.formatted_text import HTML
from prompt_toolkit import PromptSession
from janito.cli.chat_mode.toolbar import get_toolbar_func
from prompt_toolkit.enums import EditingMode
from janito.cli.chat_mode.prompt_style import chat_shell_style
from janito.cli.chat_mode.bindings import KeyBindingsFactory
from janito.cli.chat_mode.shell.commands import handle_command
from janito.cli.chat_mode.shell.autocomplete import ShellCommandCompleter


class ChatShellState:
    def __init__(self, mem_history, conversation_history):
        self.mem_history = mem_history
        self.conversation_history = conversation_history
        self.paste_mode = False
        self.termweb_port = None
        self.termweb_pid = None
        self.termweb_stdout_path = None
        self.termweb_stderr_path = None
        self.livereload_stderr_path = None
        self.termweb_status = "starting"  # Tracks the current termweb status (updated by background thread/UI)
        self.termweb_live_status = (
            None  # 'online', 'offline', updated by background checker
        )
        self.termweb_live_checked_time = None  # datetime.datetime of last status check
        self.last_usage_info = {}
        self.last_elapsed = None
        self.main_agent = {}
        self.mode = None
        self.agent = None
        self.main_agent = None
        self.main_enabled = False


class ChatSession:
    def __init__(
        self,
        console,
        provider_instance=None,
        llm_driver_config=None,
        role=None,
        args=None,
        verbose_tools=False,
        verbose_agent=False,
    ):
        from janito.cli.prompt_core import PromptHandler as GenericPromptHandler

        self._prompt_handler = GenericPromptHandler(
            args=None,
            conversation_history=(
                None
                if not hasattr(self, "shell_state")
                else self.shell_state.conversation_history
            ),
            provider_instance=provider_instance,
        )
        self._prompt_handler.agent = None  # Will be set below if agent exists
        self.console = console
        self.user_input_history = UserInputHistory()
        self.input_dicts = self.user_input_history.load()
        self.mem_history = InMemoryHistory()
        for item in self.input_dicts:
            if isinstance(item, dict) and "input" in item:
                self.mem_history.append_string(item["input"])
        self.provider_instance = provider_instance
        self.llm_driver_config = llm_driver_config
        from janito.agent.setup_agent import create_configured_agent

        agent = create_configured_agent(
            provider_instance=provider_instance,
            llm_driver_config=llm_driver_config,
            role=role,
            verbose_tools=verbose_tools,
            verbose_agent=verbose_agent,
        )
        from janito.conversation_history import LLMConversationHistory

        self.shell_state = ChatShellState(self.mem_history, LLMConversationHistory())
        self.shell_state.agent = agent
        self.agent = agent
        from janito.perf_singleton import performance_collector

        self.performance_collector = performance_collector
        self.key_bindings = KeyBindingsFactory.create()
        # Attach agent to prompt handler now that agent is initialized
        self._prompt_handler.agent = self.agent
        self._prompt_handler.conversation_history = (
            self.shell_state.conversation_history
        )

        # TERMWEB logic migrated from runner
        self.termweb_support = False
        if args and getattr(args, "web", False):
            self.termweb_support = True
            self.shell_state.termweb_support = self.termweb_support
            from janito.cli.termweb_starter import termweb_start_and_watch
            from janito.cli.config import get_termweb_port
            import threading
            from rich.console import Console

            Console().print("[yellow]Starting termweb in background...[/yellow]")
            self.termweb_lock = threading.Lock()
            termweb_thread = termweb_start_and_watch(
                self.shell_state, self.termweb_lock, get_termweb_port()
            )
            # Initial status is set to 'starting' by constructor; the watcher will update
            self.termweb_thread = termweb_thread

            # Start a background timer to update live termweb status (for UI responsiveness)
            import threading, datetime

            def update_termweb_liveness():
                while True:
                    with self.termweb_lock:
                        port = getattr(self.shell_state, "termweb_port", None)
                        if port:
                            try:
                                # is_termweb_running is removed; inline health check here:
                                try:
                                    import http.client

                                    conn = http.client.HTTPConnection(
                                        "localhost", port, timeout=0.5
                                    )
                                    conn.request("GET", "/")
                                    resp = conn.getresponse()
                                    running = resp.status == 200
                                except Exception:
                                    running = False
                                self.shell_state.termweb_live_status = (
                                    "online" if running else "offline"
                                )
                            except Exception:
                                self.shell_state.termweb_live_status = "offline"
                            self.shell_state.termweb_live_checked_time = (
                                datetime.datetime.now()
                            )
                        else:
                            self.shell_state.termweb_live_status = None
                            self.shell_state.termweb_live_checked_time = (
                                datetime.datetime.now()
                            )
                    # sleep outside lock
                    threading.Event().wait(1.0)

            self._termweb_liveness_thread = threading.Thread(
                target=update_termweb_liveness, daemon=True
            )
            self._termweb_liveness_thread.start()
            # No queue or blocking checks; UI (and timer) will observe self.shell_state fields

        else:
            self.shell_state.termweb_support = False
            self.shell_state.termweb_status = "offline"

    def run(self):
        session = self._create_prompt_session()
        self.console.print(
            "[bold green]Type /help for commands. Type /exit or press Ctrl+C to quit.[/bold green]"
        )
        self._chat_loop(session)

    def _chat_loop(self, session):
        self.msg_count = 0
        timer_started = False
        while True:
            if not timer_started:
                timer_started = True
            cmd_input = self._handle_input(session)
            if cmd_input is None:
                break
            if not cmd_input:
                continue
            if self._handle_exit_conditions(cmd_input):
                break
            if cmd_input.startswith("/"):
                handle_command(cmd_input, shell_state=self.shell_state)
                continue
            self.user_input_history.append(cmd_input)
            try:
                final_event = (
                    self._prompt_handler.agent.last_event
                    if hasattr(self._prompt_handler.agent, "last_event")
                    else None
                )
                self._prompt_handler.run_prompt(cmd_input)
                self.msg_count += 1
                # After prompt, print the stat line using the shared core function
                from janito.formatting_token import print_token_message_summary

                usage = self.performance_collector.get_last_request_usage()
                print_token_message_summary(self.console, self.msg_count, usage)
                # Print exit reason if present in the final event
                if final_event and hasattr(final_event, "metadata"):
                    exit_reason = (
                        final_event.metadata.get("exit_reason")
                        if hasattr(final_event, "metadata")
                        else None
                    )
                    if exit_reason:
                        self.console.print(
                            f"[bold yellow]Exit reason: {exit_reason}[/bold yellow]"
                        )

            except Exception as exc:
                self.console.print(f"[red]Exception in agent: {exc}[/red]")
                import traceback

                self.console.print(traceback.format_exc())

    def _create_prompt_session(self):
        return PromptSession(
            style=chat_shell_style,
            completer=ShellCommandCompleter(),
            history=self.mem_history,
            editing_mode=EditingMode.EMACS,
            key_bindings=self.key_bindings,
            bottom_toolbar=lambda: get_toolbar_func(
                self.performance_collector, 0, self.shell_state
            )(),
        )

    def _handle_input(self, session):
        injected = getattr(self.shell_state, "injected_input", None)
        if injected is not None:
            cmd_input = injected
            self.shell_state.injected_input = None
        else:
            try:
                cmd_input = session.prompt(HTML("<inputline>💬 </inputline>"))
            except (KeyboardInterrupt, EOFError):
                self._handle_exit()
                return None
        sanitized = cmd_input.strip()
        # Ensure UTF-8 validity and sanitize if needed
        try:
            # This will raise UnicodeEncodeError if not encodable
            sanitized.encode("utf-8")
        except UnicodeEncodeError:
            # Replace invalid characters
            sanitized = sanitized.encode("utf-8", errors="replace").decode("utf-8")
            self.console.print(
                "[yellow]Warning: Some characters in your input were not valid UTF-8 and have been replaced.[/yellow]"
            )
        return sanitized

    def _handle_exit(self):
        self.console.print("[bold yellow]Exiting chat. Goodbye![/bold yellow]")
        # Ensure driver thread is joined before exit
        if hasattr(self, "agent") and hasattr(self.agent, "join_driver"):
            if (
                hasattr(self.agent, "input_queue")
                and self.agent.input_queue is not None
            ):
                self.agent.input_queue.put(None)
            self.agent.join_driver()

    def _handle_exit_conditions(self, cmd_input):
        if cmd_input.lower() in ("/exit", ":q", ":quit"):
            self._handle_exit()
            return True
        return False



from janito.cli.console import shared_console
from janito.cli.chat_mode.shell.commands.base import ShellCmdHandler

class ToolsShellHandler(ShellCmdHandler):
    help_text = "List available tools"

    def run(self):
        try:
            import janito.tools  # Ensure all tools are registered
            registry = janito.tools.get_local_tools_adapter()
            tools = registry.list_tools()
            shared_console.print("Registered tools:" if tools else "No tools registered.")
            for tool in tools:
                shared_console.print(f"- {tool}")

            # Check for execution tools
            # We assume shell_state.allow_execution is set if -x is used
            allow_execution = False
            if hasattr(self, 'shell_state') and self.shell_state is not None:
                allow_execution = getattr(self.shell_state, 'allow_execution', False)

            # Find all possible execution tools (by convention: provides_execution = True)
            exec_tools = []
            for tool_instance in registry.get_tools():
                if getattr(tool_instance, 'provides_execution', False):
                    exec_tools.append(tool_instance.tool_name)

            if not allow_execution and exec_tools:
                shared_console.print("[yellow]⚠️  Warning: Execution tools (e.g., commands, code execution) are disabled. Use -x to enable them.[/yellow]")

        except Exception as e:
            shared_console.print(f"[red]Error loading tools: {e}[/red]")



from pathlib import Path
from jinja2 import Template
import importlib.resources
import sys
import warnings
from janito.tools import get_local_tools_adapter
from janito.llm.agent import LLMAgent
from janito.drivers.driver_registry import get_driver_class
from queue import Queue
from janito.platform_discovery import PlatformDiscovery


def setup_agent(
    provider_instance,
    llm_driver_config,
    role=None,
    templates_dir=None,
    zero_mode=False,
    input_queue=None,
    output_queue=None,
    verbose_tools=False,
    verbose_agent=False,
):
    """
    Creates an agent using a rendered system prompt template, passing an explicit role.
    """
    tools_provider = get_local_tools_adapter()
    tools_provider.set_verbose_tools(verbose_tools)

    if zero_mode:
        # Pass provider to agent, let agent create driver
        agent = LLMAgent(
            provider_instance,
            tools_provider,
            agent_name=role or "software developer",
            system_prompt=None,
            verbose_agent=verbose_agent,
        )
        return agent
    # Normal flow
    if templates_dir is None:
        # Set default template directory
        templates_dir = Path(__file__).parent / "templates" / "profiles"
    template_path = templates_dir / "system_prompt_template_main.txt.j2"

    template_content = None
    if template_path.exists():
        with open(template_path, "r", encoding="utf-8") as file:
            template_content = file.read()
    else:
        # Try package import fallback: janito.agent.templates.profiles.system_prompt_template_main.txt.j2
        try:
            with importlib.resources.files("janito.agent.templates.profiles").joinpath(
                "system_prompt_template_main.txt.j2"
            ).open("r", encoding="utf-8") as file:
                template_content = file.read()
        except (FileNotFoundError, ModuleNotFoundError, AttributeError):
            warnings.warn(
                f"[janito] Could not find system_prompt_template_main.txt.j2 in {template_path} nor in janito.agent.templates.profiles package."
            )
            raise FileNotFoundError(
                f"Template file not found in either {template_path} or package resource."
            )

    template = Template(template_content)
    # Prepare context for Jinja2 rendering from llm_driver_config
    # Compose context for Jinja2 rendering without using to_dict or temperature
    context = {}
    context["role"] = role or "software developer"
    # Inject current platform environment information
    pd = PlatformDiscovery()
    context["platform"] = pd.get_platform_name()
    context["python_version"] = pd.get_python_version()
    context["shell_info"] = pd.detect_shell()
    rendered_prompt = template.render(**context)
    # Create the agent as before, now passing the explicit role
    # Do NOT pass temperature; do not depend on to_dict
    agent = LLMAgent(
        provider_instance,
        tools_provider,
        agent_name=role or "software developer",
        system_prompt=rendered_prompt,
        input_queue=input_queue,
        output_queue=output_queue,
        verbose_agent=verbose_agent,
    )
    agent.template_vars["role"] = context["role"]
    return agent


def create_configured_agent(
    *,
    provider_instance=None,
    llm_driver_config=None,
    role=None,
    verbose_tools=False,
    verbose_agent=False,
    templates_dir=None,
    zero_mode=False,
):
    """
    Normalizes agent setup for all CLI modes.

    Args:
        provider_instance: Provider instance for the agent
        llm_driver_config: LLM driver configuration
        role: Optional role string
        verbose_tools: Optional, default False
        verbose_agent: Optional, default False
        templates_dir: Optional
        zero_mode: Optional, default False

    Returns:
        Configured agent instance
    """
    # If provider_instance has create_driver, wire queues (single-shot mode)
    input_queue = None
    output_queue = None
    driver = None
    if hasattr(provider_instance, "create_driver"):
        driver = provider_instance.create_driver()
        driver.start()  # Ensure the driver background thread is started
        input_queue = getattr(driver, "input_queue", None)
        output_queue = getattr(driver, "output_queue", None)

    agent = setup_agent(
        provider_instance=provider_instance,
        llm_driver_config=llm_driver_config,
        role=role,
        templates_dir=templates_dir,
        zero_mode=zero_mode,
        input_queue=input_queue,
        output_queue=output_queue,
        verbose_tools=verbose_tools,
        verbose_agent=verbose_agent,
    )
    if driver is not None:
        agent.driver = driver  # Attach driver to agent for thread management
    return agent



from janito.tools.tool_base import ToolBase
from janito.tools.tool_events import ToolCallStarted, ToolCallFinished, ToolCallError
from janito.exceptions import ToolCallException
from typing import Optional


class ToolsAdapterBase:
    """
    Composable entry point for tools management and provisioning in LLM pipelines.
    This class represents an external or plugin-based provider of tool definitions.
    Extend and customize this to load, register, or serve tool implementations dynamically.
    After refactor, also responsible for tool execution.
    """

    def __init__(
        self, tools=None, event_bus=None, allowed_tools: Optional[list] = None
    ):
        self._tools = tools or []
        self._event_bus = event_bus  # event bus can be set on all adapters
        self._allowed_tools = set(allowed_tools) if allowed_tools is not None else None
        self.verbose_tools = False

    def set_verbose_tools(self, value: bool):
        self.verbose_tools = value

    @property
    def event_bus(self):
        return self._event_bus

    @event_bus.setter
    def event_bus(self, bus):
        self._event_bus = bus

    def get_tools(self):
        """Return the list of tools managed by this provider."""
        return self._tools

    def add_tool(self, tool):
        self._tools.append(tool)

    def clear_tools(self):
        self._tools = []

    def _validate_arguments_against_schema(self, arguments: dict, schema: dict):
        properties = schema.get("properties", {})
        required = schema.get("required", [])
        missing = [field for field in required if field not in arguments]
        if missing:
            return f"Missing required argument(s): {', '.join(missing)}"
        type_map = {
            "string": str,
            "integer": int,
            "number": (int, float),
            "boolean": bool,
            "array": list,
            "object": dict,
        }
        for key, value in arguments.items():
            if key not in properties:
                continue
            expected_type = properties[key].get("type")
            if expected_type and expected_type in type_map:
                if not isinstance(value, type_map[expected_type]):
                    return f"Argument '{key}' should be of type '{expected_type}', got '{type(value).__name__}'"
        return None

    def execute(self, tool, *args, **kwargs):

        if self.verbose_tools:
            print(
                f"[tools-adapter] [execute] Executing tool: {getattr(tool, 'tool_name', repr(tool))} with args: {args}, kwargs: {kwargs}"
            )
        if isinstance(tool, ToolBase):
            tool.event_bus = self._event_bus
        result = None
        if callable(tool):
            result = tool(*args, **kwargs)
        elif hasattr(tool, "execute") and callable(getattr(tool, "execute")):
            result = tool.execute(*args, **kwargs)
        elif hasattr(tool, "run") and callable(getattr(tool, "run")):
            result = tool.run(*args, **kwargs)
        else:
            raise ValueError("Provided tool is not executable.")

        return result

    def execute_by_name(
        self, tool_name: str, *args, request_id=None, arguments=None, **kwargs
    ):
        self._check_tool_permissions(tool_name, request_id, arguments)
        tool = self.get_tool(tool_name)
        self._ensure_tool_exists(tool, tool_name, request_id, arguments)
        schema = getattr(tool, "schema", None)
        if schema and arguments is not None:
            validation_error = self._validate_arguments_against_schema(
                arguments, schema
            )
            if validation_error:
                if self._event_bus:
                    self._event_bus.publish(
                        ToolCallError(
                            tool_name=tool_name,
                            request_id=request_id,
                            error=validation_error,
                            arguments=arguments,
                        )
                    )
                return validation_error
        if self.verbose_tools:
            print(
                f"[tools-adapter] Executing tool: {tool_name} with arguments: {arguments}"
            )
        if self._event_bus:
            self._event_bus.publish(
                ToolCallStarted(
                    tool_name=tool_name, request_id=request_id, arguments=arguments
                )
            )
        try:
            result = self.execute(tool, **(arguments or {}), **kwargs)
        except Exception as e:
            self._handle_execution_error(tool_name, request_id, e, arguments)
        if self.verbose_tools:
            print(f"[tools-adapter] Tool execution finished: {tool_name} -> {result}")
        if self._event_bus:
            self._event_bus.publish(
                ToolCallFinished(
                    tool_name=tool_name, request_id=request_id, result=result
                )
            )
        return result

    def execute_function_call_message_part(self, function_call_message_part):
        """
        Execute a FunctionCallMessagePart by extracting the tool name and arguments and dispatching to execute_by_name.
        """
        import json

        function = getattr(function_call_message_part, "function", None)
        tool_call_id = getattr(function_call_message_part, "tool_call_id", None)
        if function is None or not hasattr(function, "name"):
            raise ValueError(
                "FunctionCallMessagePart does not contain a valid function object."
            )
        tool_name = function.name
        arguments = function.arguments
        # Parse arguments if they are a JSON string
        if isinstance(arguments, str):
            try:
                arguments = json.loads(arguments)
            except Exception:
                pass  # Leave as string if not JSON
        if self.verbose_tools:
            print(
                f"[tools-adapter] Executing FunctionCallMessagePart: tool={tool_name}, arguments={arguments}, tool_call_id={tool_call_id}"
            )
        return self.execute_by_name(
            tool_name, request_id=tool_call_id, arguments=arguments
        )

    def _check_tool_permissions(self, tool_name, request_id, arguments):
        if self._allowed_tools is not None and tool_name not in self._allowed_tools:
            error_msg = f"Tool '{tool_name}' is not permitted by adapter allow-list."
            if self._event_bus:
                self._event_bus.publish(
                    ToolCallError(
                        tool_name=tool_name,
                        request_id=request_id,
                        error=error_msg,
                        arguments=arguments,
                    )
                )
            raise ToolCallException(tool_name, error_msg, arguments=arguments)

    def _ensure_tool_exists(self, tool, tool_name, request_id, arguments):
        if tool is None:
            error_msg = f"Tool '{tool_name}' not found in registry."
            if self._event_bus:
                self._event_bus.publish(
                    ToolCallError(
                        tool_name=tool_name,
                        request_id=request_id,
                        error=error_msg,
                        arguments=arguments,
                    )
                )
            raise ToolCallException(tool_name, error_msg, arguments=arguments)

    def _handle_execution_error(self, tool_name, request_id, exception, arguments):
        error_msg = f"Exception during execution of tool '{tool_name}': {exception}"
        if self._event_bus:
            self._event_bus.publish(
                ToolCallError(
                    tool_name=tool_name,
                    request_id=request_id,
                    error=error_msg,
                    exception=exception,
                    arguments=arguments,
                )
            )
        raise ToolCallException(
            tool_name, error_msg, arguments=arguments, exception=exception
        )

    def get_tool(self, tool_name):
        """Abstract method: implement in subclass to return tool instance by name"""
        raise NotImplementedError()



import os
import webbrowser
from janito.cli.console import shared_console
from janito.cli.chat_mode.shell.commands.base import ShellCmdHandler


class EditShellHandler(ShellCmdHandler):
    help_text = "Open a file in the browser-based editor"

    def run(self):
        filename = self.after_cmd_line.strip()
        if not filename:
            shared_console.print("[red]Usage: /edit <filename>[/red]")
            return
        if not os.path.isfile(filename):
            shared_console.print(f"[red]File not found:[/red] {filename}")
            return
        from janito.cli.config import get_termweb_port

        port = get_termweb_port()
        url = f"http://localhost:{port}/?path={filename}"
        shared_console.print(
            f"[green]Opening in browser:[/green] [underline blue]{url}[/underline blue]"
        )
        webbrowser.open(url)



from prompt_toolkit.formatted_text import HTML
from janito.performance_collector import PerformanceCollector
from janito.cli.config import config
from janito.version import __version__ as VERSION


def format_tokens(n, tag=None):
    if n is None:
        return "?"
    if n < 1000:
        val = str(n)
    elif n < 1000000:
        val = f"{n/1000:.1f}k"
    else:
        val = f"{n/1000000:.1f}M"
    return f"<{tag}>{val}</{tag}>" if tag else val


def assemble_first_line(provider_name, model_name, role, agent=None):
    return f" Janito {VERSION} | Provider: <provider>{provider_name}</provider> | Model: <model>{model_name}</model> | Role: <role>{role}</role>"


def assemble_bindings_line(width):
    return (
        f" <key-label>F1</key-label>: Restart conversation | "
        f"<b>/help</b>: Help | "
        f"<key-label>F12</key-label>: Do It "
    )


def get_toolbar_func(perf: PerformanceCollector, msg_count: int, shell_state):
    from prompt_toolkit.application.current import get_app
    import importlib

    def get_toolbar():
        width = get_app().output.get_size().columns
        provider_name = "?"
        model_name = "?"
        role = "?"
        agent = shell_state.agent if hasattr(shell_state, "agent") else None
        termweb_support = getattr(shell_state, "termweb_support", False)
        termweb_port = (
            shell_state.termweb_port if hasattr(shell_state, "termweb_port") else None
        )
        termweb_status = (
            shell_state.termweb_status
            if hasattr(shell_state, "termweb_status")
            else None
        )
        # Use cached liveness check only (set by background thread in shell_state)
        this_termweb_status = termweb_status
        if not termweb_support:
            this_termweb_status = None
        elif termweb_status == "starting" or termweb_status is None:
            this_termweb_status = termweb_status
        else:
            live_status = (
                shell_state.termweb_live_status
                if hasattr(shell_state, "termweb_live_status")
                else None
            )
            if live_status is not None:
                this_termweb_status = live_status
        if agent is not None:
            # Use agent API to get provider and model name
            provider_name = (
                agent.get_provider_name()
                if hasattr(agent, "get_provider_name")
                else "?"
            )
            model_name = (
                agent.get_model_name() if hasattr(agent, "get_model_name") else "?"
            )
            if hasattr(agent, "template_vars"):
                role = agent.template_vars.get("role", "?")
        usage = perf.get_last_request_usage()
        first_line = assemble_first_line(provider_name, model_name, role, agent=agent)

        bindings_line = assemble_bindings_line(width)
        toolbar_text = first_line + "\n" + bindings_line
        # Add termweb status if available, after the F12 line
        if this_termweb_status == "online" and termweb_port:
            toolbar_text += f"\n<termweb> Termweb </termweb>Online<termweb> at <u>http://localhost:{termweb_port}</u></termweb>"
        elif this_termweb_status == "starting":
            toolbar_text += "\n<termweb> Termweb </termweb>Starting"
        elif this_termweb_status == "offline":
            toolbar_text += "\n<termweb> Termweb </termweb>Offline"
        return HTML(toolbar_text)

    return get_toolbar



"""
PromptHandler: Handles prompt submission and response formatting for janito CLI (one-shot prompt execution).
"""

import time
from janito.version import __version__ as VERSION
from janito.cli.prompt_core import PromptHandler as GenericPromptHandler
from janito.cli.verbose_output import (
    print_verbose_header,
    print_performance,
    handle_exception,
)
import janito.tools  # Ensure all tools are registered
from janito.cli.console import shared_console


class PromptHandler:
    def __init__(self, args, provider_instance, llm_driver_config, role=None):
        self.args = args
        self.provider_instance = provider_instance
        self.llm_driver_config = llm_driver_config
        self.role = role
        from janito.agent.setup_agent import create_configured_agent

        self.agent = create_configured_agent(
            provider_instance=provider_instance,
            llm_driver_config=llm_driver_config,
            role=role,
            verbose_tools=getattr(args, "verbose_tools", False),
            verbose_agent=getattr(args, "verbose_agent", False),
        )
        # Setup conversation/history if needed
        self.generic_handler = GenericPromptHandler(
            args, [], provider_instance=provider_instance
        )
        self.generic_handler.agent = self.agent

    def handle(self) -> None:
        import traceback

        user_prompt = " ".join(getattr(self.args, "user_prompt", [])).strip()
        # UTF-8 sanitize user_prompt
        sanitized = user_prompt
        try:
            sanitized.encode("utf-8")
        except UnicodeEncodeError:
            sanitized = sanitized.encode("utf-8", errors="replace").decode("utf-8")
            shared_console.print(
                "[yellow]Warning: Some characters in your input were not valid UTF-8 and have been replaced.[/yellow]"
            )
        try:
            self.generic_handler.handle_prompt(
                sanitized,
                args=self.args,
                print_header=True,
                raw=getattr(self.args, "raw", False),
            )
            if hasattr(self.args, "verbose_agent") and self.args.verbose_agent:
                print("[debug] handle_prompt() completed without exception.")
        except Exception as e:
            print(
                f"[error] Exception occurred in handle_prompt: {type(e).__name__}: {e}"
            )
            traceback.print_exc()
        self._post_prompt_actions()

    def _post_prompt_actions(self):
        final_event = getattr(self.agent, "last_event", None)
        if final_event is not None:
            self._print_exit_reason_and_parts(final_event)
            # --- BEGIN: Print token info in rich rule if --verbose is set ---
            if hasattr(self.args, "verbose") and self.args.verbose:
                from janito.perf_singleton import performance_collector

                token_info = performance_collector.get_last_request_usage()
                from rich.rule import Rule
                from rich import print as rich_print
                from janito.cli.utils import format_tokens

                if token_info:
                    if isinstance(token_info, dict):
                        token_str = " | ".join(
                            f"{k}: {format_tokens(v) if isinstance(v, int) else v}"
                            for k, v in token_info.items()
                        )
                    else:
                        token_str = str(token_info)
                    rich_print(Rule(f"[bold cyan]Token Usage[/bold cyan] {token_str}"))
                else:
                    rich_print(Rule("[cyan]No token usage info available.[/cyan]"))
        else:
            shared_console.print("[yellow]No output produced by the model.[/yellow]")
        self._cleanup_driver_and_console()

    def _print_exit_reason_and_parts(self, final_event):
        exit_reason = (
            getattr(final_event, "metadata", {}).get("exit_reason")
            if hasattr(final_event, "metadata")
            else None
        )
        if exit_reason:
            print(f"[bold yellow]Exit reason: {exit_reason}[/bold yellow]")
        parts = getattr(final_event, "parts", None)
        if not exit_reason:
            if parts is None or len(parts) == 0:
                shared_console.print(
                    "[yellow]No output produced by the model.[/yellow]"
                )
            else:
                if hasattr(self.args, "verbose_agent") and self.args.verbose_agent:
                    print(
                        "[yellow]No user-visible output. Model returned the following parts:"
                    )
                    for idx, part in enumerate(parts):
                        print(
                            f"  [part {idx}] type: {type(part).__name__} | content: {getattr(part, 'content', repr(part))}"
                        )

    def _cleanup_driver_and_console(self):
        if hasattr(self.agent, "join_driver"):
            if (
                hasattr(self.agent, "input_queue")
                and self.agent.input_queue is not None
            ):
                self.agent.input_queue.put(None)
            self.agent.join_driver()
        try:
            shared_console.file.flush()
        except Exception:
            pass
        try:
            import sys

            sys.stdout.flush()
        except Exception:
            pass
        # If event logger is active, flush event log



import inspect
import typing
from janito.tools.tools_schema import ToolSchemaBase

try:
    from google.genai import types as genai_types
except ImportError:
    genai_types = None


class GeminiSchemaGenerator(ToolSchemaBase):
    PYTHON_TYPE_TO_GENAI_TYPE = {
        str: "STRING",
        int: "INTEGER",
        float: "NUMBER",
        bool: "BOOLEAN",
        list: "ARRAY",
        dict: "OBJECT",
    }

    def type_to_genai_schema(self, annotation, description=None):
        if hasattr(annotation, "__origin__"):
            if annotation.__origin__ is list or annotation.__origin__ is typing.List:
                return genai_types.Schema(
                    type="ARRAY",
                    items=self.type_to_genai_schema(annotation.__args__[0]),
                    description=description,
                )
            if annotation.__origin__ is dict or annotation.__origin__ is typing.Dict:
                return genai_types.Schema(type="OBJECT", description=description)
        return genai_types.Schema(
            type=self.PYTHON_TYPE_TO_GENAI_TYPE.get(annotation, "STRING"),
            description=description,
        )

    def generate_declaration(self, tool_class):
        func, tool_name, sig, summary, param_descs, return_desc, description = (
            self.validate_tool_class(tool_class)
        )
        properties = {}
        required = []
        # Removed tool_call_reason from properties and required
        for name, param in sig.parameters.items():
            if name == "self":
                continue
            annotation = param.annotation
            pdesc = param_descs.get(name, "")
            schema = self.type_to_genai_schema(annotation, description=pdesc)
            properties[name] = schema
            if param.default == inspect._empty:
                required.append(name)
        parameters_schema = genai_types.Schema(
            type="OBJECT", properties=properties, required=required
        )
        return genai_types.FunctionDeclaration(
            name=tool_name, description=description, parameters=parameters_schema
        )


def generate_tool_declarations(tool_classes: list):
    if genai_types is None:
        raise ImportError("google-genai package is not installed.")
    generator = GeminiSchemaGenerator()
    function_declarations = [
        generator.generate_declaration(tool_class) for tool_class in tool_classes
    ]
    return [genai_types.Tool(function_declarations=function_declarations)]



"""
ProviderRegistry: Handles provider listing and selection logic for janito CLI.
"""

from rich.table import Table
from janito.cli.console import shared_console
from janito.providers.registry import LLMProviderRegistry
from janito.providers.provider_static_info import STATIC_PROVIDER_METADATA
from janito.llm.auth import LLMAuthManager
import sys
from janito.exceptions import MissingProviderSelectionException


class ProviderRegistry:
    def list_providers(self):
        """List all supported LLM providers as a table using rich, showing if auth is configured and supported model names."""
        providers = self._get_provider_names()
        table = self._create_table()
        rows = self._get_all_provider_rows(providers)
        self._add_rows_to_table(table, rows)
        self._print_table(table)

    def _get_provider_names(self):
        return list(STATIC_PROVIDER_METADATA.keys())

    def _create_table(self):
        table = Table(title="Supported LLM Providers")
        table.add_column("Provider", style="cyan")
        table.add_column("Maintainer", style="yellow", justify="center")
        table.add_column("Model Names", style="magenta")
        return table

    def _get_all_provider_rows(self, providers):
        rows = []
        for p in providers:
            info = self._get_provider_info(p)
            # info is (provider_name, maintainer, model_names, skip)
            if len(info) == 4 and info[3]:
                continue  # skip providers flagged as not implemented
            rows.append(info[:3])
        rows.sort(key=self._maintainer_sort_key)
        return rows

    def _add_rows_to_table(self, table, rows):
        for idx, (p, maintainer, model_names) in enumerate(rows):
            table.add_row(p, maintainer, model_names)
            if idx != len(rows) - 1:
                table.add_section()

    def _print_table(self, table):
        shared_console.print(table)

    def _get_provider_info(self, provider_name):
        static_info = STATIC_PROVIDER_METADATA.get(provider_name, {})
        maintainer_val = static_info.get("maintainer", "-")
        maintainer = (
            "[red]🚨 Needs maintainer[/red]"
            if maintainer_val == "Needs maintainer"
            else f"👤 {maintainer_val}"
        )
        model_names = "-"
        unavailable_reason = None
        skip = False
        try:
            provider_class = LLMProviderRegistry.get(provider_name)
            creds = LLMAuthManager().get_credentials(provider_name)
            provider_instance = None
            instantiation_failed = False
            try:
                provider_instance = provider_class()
            except NotImplementedError:
                skip = True
                unavailable_reason = "Not implemented"
                model_names = f"[red]❌ Not implemented[/red]"
            except Exception as e:
                instantiation_failed = True
                unavailable_reason = (
                    f"Unavailable (import error or missing dependency): {str(e)}"
                )
                model_names = f"[red]❌ {unavailable_reason}[/red]"
            if not instantiation_failed and provider_instance is not None:
                available, unavailable_reason = self._get_availability(
                    provider_instance
                )
                if (
                    not available
                    and unavailable_reason
                    and "not implemented" in str(unavailable_reason).lower()
                ):
                    skip = True
                if available:
                    model_names = self._get_model_names(provider_name)
                else:
                    model_names = f"[red]❌ {unavailable_reason}[/red]"
        except Exception as import_error:
            model_names = f"[red]❌ Unavailable (cannot import provider module): {str(import_error)}[/red]"
        return (provider_name, maintainer, model_names, skip)

    def _get_availability(self, provider_instance):
        try:
            available = getattr(provider_instance, "available", True)
            unavailable_reason = getattr(provider_instance, "unavailable_reason", None)
        except Exception as e:
            available = False
            unavailable_reason = f"Error reading runtime availability: {str(e)}"
        return available, unavailable_reason

    def _get_model_names(self, provider_name):
        provider_to_specs = {
            "openai": "janito.providers.openai.model_info",
            "azure_openai": "janito.providers.azure_openai.model_info",
            "google": "janito.providers.google.model_info",
            "mistralai": "janito.providers.mistralai.model_info",
            "deepseek": "janito.providers.deepseek.model_info",
        }
        if provider_name in provider_to_specs:
            try:
                mod = __import__(
                    provider_to_specs[provider_name], fromlist=["MODEL_SPECS"]
                )
                return ", ".join(mod.MODEL_SPECS.keys())
            except Exception:
                return "(Error)"
        return "-"

    def _maintainer_sort_key(self, row):
        maint = row[1]
        is_needs_maint = "Needs maintainer" in maint
        return (is_needs_maint, row[2] != "✅ Auth")

    def get_provider(self, provider_name):
        """Return the provider class for the given provider name."""
        from janito.providers.registry import LLMProviderRegistry

        if not provider_name:
            raise ValueError("Provider name must be specified.")
        return LLMProviderRegistry.get(provider_name)

    def get_instance(self, provider_name, config=None):
        """Return an instance of the provider for the given provider name, optionally passing a config object."""
        provider_class = self.get_provider(provider_name)
        if provider_class is None:
            raise ValueError(f"No provider class found for '{provider_name}'")
        if config is not None:
            return provider_class(config=config)
        return provider_class()


# For backward compatibility
def list_providers():
    """Legacy function for listing providers, now uses ProviderRegistry class."""
    ProviderRegistry().list_providers()



import os
from janito.cli.chat_mode.shell.session.manager import reset_session_id
from janito.cli.chat_mode.shell.commands.base import ShellCmdHandler
from janito.cli.console import shared_console


def handle_restart(shell_state=None):
    from janito.cli.chat_mode.shell.session.manager import (
        load_last_conversation,
        save_conversation,
    )

    reset_session_id()
    save_path = os.path.join(".janito", "last_conversation.json")

    # --- Append end-of-conversation message to old history if it exists and is non-trivial ---
    if os.path.exists(save_path):
        try:
            messages, prompts, usage = load_last_conversation(save_path)
            if messages and (
                len(messages) > 1
                or (len(messages) == 1 and messages[0].get("role") != "system")
            ):
                messages.append(
                    {"role": "system", "content": "[Session ended by user]"}
                )
                # Save to permanent chat history (let save_conversation pick session file)
                save_conversation(messages, prompts, usage)
        except Exception as e:
            shared_console.print(
                f"[bold red]Failed to update previous conversation history:[/bold red] {e}"
            )

    # Clear the terminal screen
    shared_console.clear()

    # Reset conversation history using the agent's method
    if hasattr(shell_state, "agent") and shell_state.agent:
        shell_state.agent.reset_conversation_history()

    # Reset tool use tracker
    try:
        from janito.tools.tool_use_tracker import ToolUseTracker

        ToolUseTracker.instance().clear_history()
    except Exception as e:
        shared_console.print(
            f"[bold yellow]Warning: Failed to reset tool use tracker:[/bold yellow] {e}"
        )

    # Reset token usage info in-place so all references (including status bar) are updated
    for k in ("prompt_tokens", "completion_tokens", "total_tokens"):
        shell_state.last_usage_info[k] = 0
    shell_state.last_elapsed = None

    # Reset the performance collector's last usage (so toolbar immediately reflects cleared stats)
    try:
        from janito.perf_singleton import performance_collector

        performance_collector.reset_last_request_usage()
    except Exception as e:
        shared_console.print(
            f"[bold yellow]Warning: Failed to reset PerformanceCollector token info:[/bold yellow] {e}"
        )

    shared_console.print(
        "[bold green]Conversation history has been started (context reset).[/bold green]"
    )


handle_restart.help_text = "Start a new conversation (reset context)"


class RestartShellHandler(ShellCmdHandler):
    help_text = "Start a new conversation (reset context)"

    def run(self):
        handle_restart(self.shell_state)



"""
CLI Command: Show the resolved system prompt for the main agent (single-shot mode)
"""

from janito.cli.core.runner import prepare_llm_driver_config
from janito.platform_discovery import PlatformDiscovery
from pathlib import Path
from jinja2 import Template
import importlib.resources


def handle_show_system_prompt(args):
    # Collect modifiers as in JanitoCLI
    from janito.cli.main_cli import MODIFIER_KEYS

    modifiers = {
        k: getattr(args, k) for k in MODIFIER_KEYS if getattr(args, k, None) is not None
    }
    provider, llm_driver_config, agent_role = prepare_llm_driver_config(args, modifiers)
    if provider is None or llm_driver_config is None:
        print("Error: Could not resolve provider or LLM driver config.")
        return

    # Prepare context for Jinja2 rendering
    context = {}
    context["role"] = agent_role or "software developer"
    pd = PlatformDiscovery()
    context["platform"] = pd.get_platform_name()
    context["python_version"] = pd.get_python_version()
    context["shell_info"] = pd.detect_shell()

    # Locate and load the system prompt template
    templates_dir = (
        Path(__file__).parent.parent.parent / "agent" / "templates" / "profiles"
    )
    template_path = templates_dir / "system_prompt_template_main.txt.j2"
    template_content = None
    if template_path.exists():
        with open(template_path, "r", encoding="utf-8") as file:
            template_content = file.read()
    else:
        # Try package import fallback
        try:
            with importlib.resources.files("janito.agent.templates.profiles").joinpath(
                "system_prompt_template_main.txt.j2"
            ).open("r", encoding="utf-8") as file:
                template_content = file.read()
        except (FileNotFoundError, ModuleNotFoundError, AttributeError):
            print(
                f"[janito] Could not find system_prompt_template_main.txt.j2 in {template_path} nor in janito.agent.templates.profiles package."
            )
            print("No system prompt is set or resolved for this configuration.")
            return

    template = Template(template_content)
    system_prompt = template.render(**context)

    print("\n--- System Prompt (resolved) ---\n")
    print(system_prompt)
    print("\n-------------------------------\n")
    if agent_role:
        print(f"[Role: {agent_role}]")



from janito.tools.tool_base import ToolBase
from janito.tools.adapters.local.adapter import register_local_tool
from janito.report_events import ReportAction
import webbrowser
import os

@register_local_tool
class OpenHtmlInBrowserTool(ToolBase):
    """
    Opens an HTML file in the default web browser.

    Args:
        file_path (str): The path to the HTML file to open.

    Returns:
        str: Status message indicating the result. Example:
            - "✅ Successfully opened the file in the default browser."
            - "⚠️ Error: The specified file does not exist."
            - "⚠️ Error: The specified file is not an HTML file."
    """

    tool_name = "open_html_in_browser"

    def run(self, file_path: str) -> str:
        from janito.i18n import tr
        disp_path = file_path
        self.report_action(tr("📖 Opening HTML file: '{disp_path}'", disp_path=disp_path), ReportAction.EXECUTE)

        if not os.path.exists(file_path):
            self.report_error(tr("⚠️ The specified file does not exist: '{disp_path}'", disp_path=disp_path))
            return "⚠️ The specified file does not exist."

        if not file_path.lower().endswith('.html'):
            self.report_error(tr("⚠️ The specified file is not an HTML file: '{disp_path}'", disp_path=disp_path))
            return "⚠️ The specified file is not an HTML file."

        try:
            webbrowser.open(f"file://{os.path.abspath(file_path)}")
            self.report_success(tr("✅ Ok"))
            return "✅ Ok"
        except Exception as e:
            self.report_error(tr("⚠️ Failed to open the HTML file: {err}", err=str(e)))
            return f"⚠️ Failed to open the file: {str(e)}"



from janito.llm.model import LLMModelInfo

MODEL_SPECS = {
    "gemini-2.5-pro-preview-05-06": LLMModelInfo(
        name="gemini-2.5-pro-preview-05-06",
        context=131072,
        max_input=131072,
        max_cot="N/A",
        max_response=32768,
        thinking_supported=True,
        default_temp=0.2,
        open="google",
        driver="GoogleGenaiModelDriver",
        other={"preview": True},
    ),
    "gemini-2.5-flash-preview-05-20": LLMModelInfo(
        name="gemini-2.5-flash-preview-05-20",
        context=1000000,
        max_input=1000000,
        max_cot="N/A",
        max_response=65536,
        thinking_supported=True,
        default_temp=0.2,
        open="google",
        driver="GoogleGenaiModelDriver",
        other={"preview": True, "flash": True},
    ),
    "gemini-2.5-flash-preview-04-17": LLMModelInfo(
        name="gemini-2.5-flash-preview-04-17",
        context=1000000,
        max_input=1000000,
        max_cot="N/A",
        max_response=65536,
        thinking_supported=True,
        default_temp=0.2,
        open="google",
        driver="GoogleGenaiModelDriver",
        other={"preview": True, "flash": True},
    ),
}



from janito.cli.chat_mode.shell.commands.base import ShellCmdHandler
from janito.cli.console import shared_console


class HelpShellHandler(ShellCmdHandler):
    help_text = "Show this help message"

    def run(self):
        from . import (
            COMMAND_HANDLERS,
        )  # Import moved inside method to avoid circular import

        shared_console.print("[bold magenta]Available commands:[/bold magenta]")
        for cmd, handler_cls in sorted(COMMAND_HANDLERS.items()):
            help_text = getattr(handler_cls, "help_text", "")
            shared_console.print(f"[cyan]{cmd}[/cyan]: {help_text}")



"""
Google Gemini LLM driver.

This driver handles interaction with the Google Gemini API, including support for tool/function calls and event publishing.
"""

import json
import time
import uuid
import traceback
from typing import Optional, List, Dict, Any, Union
from janito.llm.driver import LLMDriver
from janito.drivers.google_genai.schema_generator import generate_tool_declarations
from janito.driver_events import (
    GenerationStarted,
    GenerationFinished,
    RequestStarted,
    RequestFinished,
    ResponseReceived,
    RequestStatus,
)
from janito.tools.adapters.local.adapter import LocalToolsAdapter
from janito.llm.message_parts import TextMessagePart, FunctionCallMessagePart
from janito.llm.driver_config import LLMDriverConfig


def extract_usage_metadata_native(usage_obj):
    if usage_obj is None:
        return {}
    result = {}
    for attr in dir(usage_obj):
        if attr.startswith("_") or attr == "__class__":
            continue
        value = getattr(usage_obj, attr)
        if isinstance(value, (str, int, float, bool, type(None))):
            result[attr] = value
        elif isinstance(value, list):
            if all(isinstance(i, (str, int, float, bool, type(None))) for i in value):
                result[attr] = value
    return result


class GoogleGenaiModelDriver(LLMDriver):
    available = False
    unavailable_reason = "GoogleGenaiModelDriver is not implemented yet."

    @classmethod
    def is_available(cls):
        return cls.available

    name = "google_genai"

    def __init__(self, tools_adapter=None):
        raise ImportError(self.unavailable_reason)



"""
CLI Command: List available tools
"""


def handle_list_tools(args=None):
    from janito.tools.adapters.local.adapter import LocalToolsAdapter
    import janito.tools  # Ensure all tools are registered

    registry = janito.tools.get_local_tools_adapter()
    tools = registry.list_tools()
    if tools:
        from rich.table import Table
        from rich.console import Console
        console = Console()
        # Get tool instances to check provides_execution and get info
        tool_instances = {t.tool_name: t for t in registry.get_tools()}
        normal_tools = []
        exec_tools = []
        for tool in tools:
            inst = tool_instances.get(tool, None)
            # Extract parameter names from run signature
            param_names = []
            if inst and hasattr(inst, "run"):
                import inspect
                sig = inspect.signature(inst.run)
                param_names = [p for p in sig.parameters if p != "self"]
            info = {
                "name": tool,
                "params": ", ".join(param_names),
            }
            if getattr(inst, "provides_execution", False):
                exec_tools.append(info)
            else:
                normal_tools.append(info)
        table = Table(title="Registered tools", show_header=True, header_style="bold", show_lines=False, box=None)
        table.add_column("Name", style="cyan", no_wrap=True)
        table.add_column("Parameters", style="yellow")
        for info in normal_tools:
            table.add_row(info["name"], info["params"] or "-")
        console.print(table)
        if exec_tools:
            exec_table = Table(title="Execution tools (only available with -x)", show_header=True, header_style="bold", show_lines=False, box=None)
            exec_table.add_column("Name", style="cyan", no_wrap=True)
            exec_table.add_column("Parameters", style="yellow")
            for info in exec_tools:
                exec_table.add_row(info["name"], info["params"] or "-")
            console.print(exec_table)
    else:
        print("No tools registered.")
    import sys

    sys.exit(0)



import uuid
import traceback
from rich import pretty
from janito.llm.driver import LLMDriver
from janito.llm.driver_input import DriverInput
from janito.driver_events import RequestFinished, RequestStatus

# Safe import of openai SDK
try:
    import openai

    DRIVER_AVAILABLE = True
    DRIVER_UNAVAILABLE_REASON = None
except ImportError:
    DRIVER_AVAILABLE = False
    DRIVER_UNAVAILABLE_REASON = "Missing dependency: openai (pip install openai)"


class OpenAIModelDriver(LLMDriver):
    def _get_message_from_result(self, result):
        """Extract the message object from the provider result (OpenAI-specific)."""
        if hasattr(result, "choices") and result.choices:
            return result.choices[0].message
        return None

    """
    OpenAI LLM driver (threaded, queue-based, stateless). Uses input/output queues accessible via instance attributes.
    """
    available = DRIVER_AVAILABLE
    unavailable_reason = DRIVER_UNAVAILABLE_REASON

    def __init__(self, tools_adapter=None, provider_name=None):
        super().__init__(tools_adapter=tools_adapter, provider_name=provider_name)

    def _prepare_api_kwargs(self, config, conversation):
        """
        Prepares API kwargs for OpenAI, including tool schemas if tools_adapter is present,
        and OpenAI-specific arguments (model, max_tokens, temperature, etc.).
        """
        api_kwargs = {}
        # Tool schemas (moved from base)
        if self.tools_adapter:
            try:
                from janito.providers.openai.schema_generator import (
                    generate_tool_schemas,
                )

                tool_classes = self.tools_adapter.get_tool_classes()
                tool_schemas = generate_tool_schemas(tool_classes)
                api_kwargs["tools"] = tool_schemas
            except Exception as e:
                api_kwargs["tools"] = []
                if hasattr(config, "verbose_api") and config.verbose_api:
                    print(f"[OpenAIModelDriver] Tool schema generation failed: {e}")
        # OpenAI-specific parameters
        if config.model:
            api_kwargs["model"] = config.model
        # Prefer max_completion_tokens if present, else fallback to max_tokens (for backward compatibility)
        if (
            hasattr(config, "max_completion_tokens")
            and config.max_completion_tokens is not None
        ):
            api_kwargs["max_completion_tokens"] = int(config.max_completion_tokens)
        elif hasattr(config, "max_tokens") and config.max_tokens is not None:
            # For models that do not support 'max_tokens', map to 'max_completion_tokens'
            api_kwargs["max_completion_tokens"] = int(config.max_tokens)
        for p in (
            "temperature",
            "top_p",
            "presence_penalty",
            "frequency_penalty",
            "stop",
        ):
            v = getattr(config, p, None)
            if v is not None:
                api_kwargs[p] = v
        api_kwargs["messages"] = conversation
        api_kwargs["stream"] = False
        return api_kwargs

    def _call_api(self, driver_input: DriverInput):
        cancel_event = getattr(driver_input, "cancel_event", None)
        config = driver_input.config
        conversation = self.convert_history_to_api_messages(
            driver_input.conversation_history
        )
        request_id = getattr(config, "request_id", None)
        if config.verbose_api:
            print(
                f"[verbose-api] OpenAI API call about to be sent. Model: {config.model}, max_tokens: {config.max_tokens}, tools_adapter: {type(self.tools_adapter).__name__ if self.tools_adapter else None}",
                flush=True,
            )
        try:
            client = self._instantiate_openai_client(config)
            api_kwargs = self._prepare_api_kwargs(config, conversation)
            if config.verbose_api:
                print(
                    f"[OpenAI] API CALL: chat.completions.create(**{api_kwargs})",
                    flush=True,
                )
            if self._check_cancel(cancel_event, request_id, before_call=True):
                return None
            result = client.chat.completions.create(**api_kwargs)
            if self._check_cancel(cancel_event, request_id, before_call=False):
                return None
            self._print_verbose_result(config, result)
            usage_dict = self._extract_usage(result)
            if config.verbose_api:
                print(
                    f"[OpenAI][DEBUG] Attaching usage info to RequestFinished: {usage_dict}",
                    flush=True,
                )
            self.output_queue.put(
                RequestFinished(
                    driver_name=self.__class__.__name__,
                    request_id=request_id,
                    response=result,
                    status=RequestStatus.SUCCESS,
                    usage=usage_dict,
                )
            )
            if config.verbose_api:
                pretty.install()
                print("[OpenAI] API RESPONSE:", flush=True)
                pretty.pprint(result)
            return result
        except Exception as e:
            print(f"[ERROR] Exception during OpenAI API call: {e}", flush=True)
            print(f"[ERROR] config: {config}", flush=True)
            print(
                f"[ERROR] api_kwargs: {api_kwargs if 'api_kwargs' in locals() else 'N/A'}",
                flush=True,
            )
            import traceback

            print("[ERROR] Full stack trace:", flush=True)
            print(traceback.format_exc(), flush=True)
            raise

    def _instantiate_openai_client(self, config):
        try:
            api_key_display = str(config.api_key)
            if api_key_display and len(api_key_display) > 8:
                api_key_display = api_key_display[:4] + "..." + api_key_display[-4:]
            client_kwargs = {"api_key": config.api_key}
            if getattr(config, "base_url", None):
                client_kwargs["base_url"] = config.base_url
            client = openai.OpenAI(**client_kwargs)
            return client
        except Exception as e:
            print(
                f"[ERROR] Exception during OpenAI client instantiation: {e}", flush=True
            )
            import traceback

            print(traceback.format_exc(), flush=True)
            raise

    def _check_cancel(self, cancel_event, request_id, before_call=True):
        if cancel_event is not None and cancel_event.is_set():
            status = RequestStatus.CANCELLED
            reason = (
                "Cancelled before API call"
                if before_call
                else "Cancelled during API call"
            )
            self.output_queue.put(
                RequestFinished(
                    driver_name=self.__class__.__name__,
                    request_id=request_id,
                    status=status,
                    reason=reason,
                )
            )
            return True
        return False

    def _print_verbose_result(self, config, result):
        if config.verbose_api:
            print("[OpenAI] API RAW RESULT:", flush=True)
            pretty.pprint(result)
            if hasattr(result, "__dict__"):
                print("[OpenAI] API RESULT __dict__:", flush=True)
                pretty.pprint(result.__dict__)
            try:
                print("[OpenAI] API RESULT as dict:", dict(result), flush=True)
            except Exception:
                pass
            print(
                f"[OpenAI] API RESULT .usage: {getattr(result, 'usage', None)}",
                flush=True,
            )
            try:
                print(f"[OpenAI] API RESULT ['usage']: {result['usage']}", flush=True)
            except Exception:
                pass
            if not hasattr(result, "usage") or getattr(result, "usage", None) is None:
                print(
                    "[OpenAI][WARNING] No usage info found in API response.", flush=True
                )

    def _extract_usage(self, result):
        usage = getattr(result, "usage", None)
        if usage is not None:
            usage_dict = self._usage_to_dict(usage)
            if usage_dict is None:
                print(
                    "[OpenAI][WARNING] Could not convert usage to dict, using string fallback.",
                    flush=True,
                )
                usage_dict = str(usage)
        else:
            usage_dict = self._extract_usage_from_result_dict(result)
        return usage_dict

    def _usage_to_dict(self, usage):
        if hasattr(usage, "model_dump") and callable(getattr(usage, "model_dump")):
            try:
                return usage.model_dump()
            except Exception:
                pass
        if hasattr(usage, "dict") and callable(getattr(usage, "dict")):
            try:
                return usage.dict()
            except Exception:
                pass
        try:
            return dict(usage)
        except Exception:
            try:
                return vars(usage)
            except Exception:
                pass
        return None

    def _extract_usage_from_result_dict(self, result):
        try:
            return result["usage"]
        except Exception:
            return None

    def convert_history_to_api_messages(self, conversation_history):
        """
        Convert LLMConversationHistory to the list of dicts required by OpenAI's API.
        Handles 'tool_results' and 'tool_calls' roles for compliance.
        """
        import json

        api_messages = []
        for msg in conversation_history.get_history():
            role = msg.get("role")
            content = msg.get("content")
            if role == "tool_results":
                # Expect content to be a list of tool result dicts or a stringified list
                try:
                    results = (
                        json.loads(content) if isinstance(content, str) else content
                    )
                except Exception:
                    results = [content]
                for result in results:
                    # result should be a dict with keys: name, content, tool_call_id
                    if isinstance(result, dict):
                        api_messages.append(
                            {
                                "role": "tool",
                                "content": result.get("content", ""),
                                "name": result.get("name", ""),
                                "tool_call_id": result.get("tool_call_id", ""),
                            }
                        )
                    else:
                        api_messages.append(
                            {
                                "role": "tool",
                                "content": str(result),
                                "name": "",
                                "tool_call_id": "",
                            }
                        )
            elif role == "tool_calls":
                # Convert to assistant message with tool_calls field
                import json

                try:
                    tool_calls = (
                        json.loads(content) if isinstance(content, str) else content
                    )
                except Exception:
                    tool_calls = []
                api_messages.append(
                    {"role": "assistant", "content": None, "tool_calls": tool_calls}
                )
            else:
                # Special handling for 'function' role: extract 'name' from metadata if present
                if role == "function":
                    name = ""
                    if isinstance(msg, dict):
                        metadata = msg.get("metadata", {})
                        name = (
                            metadata.get("name", "")
                            if isinstance(metadata, dict)
                            else ""
                        )
                    api_messages.append(
                        {"role": "tool", "content": content, "name": name}
                    )
                else:
                    api_messages.append(msg)
        return api_messages

    def _convert_completion_message_to_parts(self, message):
        """
        Convert an OpenAI completion message object to a list of MessagePart objects.
        Handles text, tool calls, and can be extended for other types.
        """
        from janito.llm.message_parts import TextMessagePart, FunctionCallMessagePart

        parts = []
        # Text content
        content = getattr(message, "content", None)
        if content:
            parts.append(TextMessagePart(content=content))
        # Tool calls
        tool_calls = getattr(message, "tool_calls", None) or []
        for tool_call in tool_calls:
            parts.append(
                FunctionCallMessagePart(
                    tool_call_id=getattr(tool_call, "id", ""),
                    function=getattr(tool_call, "function", None),
                )
            )
        # Extend here for other message part types if needed
        return parts



from rich.console import Console
from rich.markdown import Markdown
from rich.pretty import Pretty
from rich.panel import Panel
from rich.text import Text
from janito.event_bus.handler import EventHandlerBase
import janito.driver_events as driver_events
from janito.report_events import ReportSubtype, ReportAction
from janito.event_bus.bus import event_bus
from janito.llm import message_parts


class RichTerminalReporter(EventHandlerBase):
    """
    Handles UI rendering for janito events using Rich.

    - For ResponseReceived events, iterates over the 'parts' field and displays each part appropriately:
        - TextMessagePart: rendered as Markdown (uses 'content' field)
        - Other MessageParts: displayed using Pretty or a suitable Rich representation
    - For RequestFinished events, output is printed only if raw mode is enabled (using Pretty formatting).
    - Report events (info, success, error, etc.) are always printed with appropriate styling.
    """

    def __init__(self, raw_mode=False):
        from janito.cli.console import shared_console

        self.console = shared_console
        self.raw_mode = raw_mode
        import janito.report_events as report_events

        super().__init__(driver_events, report_events)
        self._waiting_printed = False

    def on_RequestStarted(self, event):
        # Print waiting message with provider name
        provider = None
        if hasattr(event, "payload") and isinstance(event.payload, dict):
            provider = event.payload.get("provider_name")
        if not provider:
            provider = getattr(event, "provider_name", None)
        if not provider:
            provider = getattr(event, "driver_name", None)
        if not provider:
            provider = "LLM"
        self.console.print(f"[bold cyan]Waiting for {provider}...[/bold cyan]", end="")

    def on_ResponseReceived(self, event):
        parts = event.parts if hasattr(event, "parts") else None
        if not parts:
            self.console.print("[No response parts to display]")
            self.console.file.flush()
            return
        for part in parts:
            if isinstance(part, message_parts.TextMessagePart):
                self.console.print(Markdown(part.content))
                self.console.file.flush()

    def on_RequestFinished(self, event):
        self.console.print("")  # Print end of line after waiting message
        self._waiting_printed = False
        response = event.response if hasattr(event, "response") else None
        if response is not None:
            if self.raw_mode:
                self.console.print(Pretty(response, expand_all=True))
                self.console.file.flush()
            # Check for 'code' and 'event' fields in the response
            code = None
            event_field = None
            if isinstance(response, dict):
                code = response.get("code")
                event_field = response.get("event")
            if event_field is not None:
                self.console.print(f"[bold yellow]Event:[/] {event_field}")
                self.console.file.flush()
        # No output if not raw_mode or if response is None

    def on_ReportEvent(self, event):
        msg = event.message if hasattr(event, "message") else None
        subtype = event.subtype if hasattr(event, "subtype") else None
        if not msg or not subtype:
            return
        if subtype == ReportSubtype.ACTION_INFO:
            if getattr(event, "action", None) in (
                getattr(ReportAction, "UPDATE", None),
                getattr(ReportAction, "WRITE", None),
                getattr(ReportAction, "DELETE", None),
            ):
                self.console.print(f"[magenta]{msg}[/magenta]", end="")
            else:
                self.console.print(msg, end="")
            self.console.file.flush()
        elif subtype in (
            ReportSubtype.SUCCESS,
            ReportSubtype.ERROR,
            ReportSubtype.WARNING,
        ):
            self.console.print(msg)
            self.console.file.flush()
        else:
            self.console.print(msg)
            self.console.file.flush()



from janito.drivers.openai.driver import OpenAIModelDriver

# Safe import of AzureOpenAI SDK
try:
    from openai import AzureOpenAI

    DRIVER_AVAILABLE = True
    DRIVER_UNAVAILABLE_REASON = None
except ImportError:
    DRIVER_AVAILABLE = False
    DRIVER_UNAVAILABLE_REASON = "Missing dependency: openai (pip install openai)"

from janito.llm.driver_config import LLMDriverConfig


class AzureOpenAIModelDriver(OpenAIModelDriver):
    available = DRIVER_AVAILABLE
    unavailable_reason = DRIVER_UNAVAILABLE_REASON

    @classmethod
    def is_available(cls):
        return cls.available

    required_config = {"base_url"}  # Update key as used in your config logic

    def __init__(self, tools_adapter=None):
        if not self.available:
            raise ImportError(
                f"AzureOpenAIModelDriver unavailable: {self.unavailable_reason}"
            )
        # Do NOT call super().__init__ if Azure SDK is not available
        OpenAIModelDriver.__init__(self, tools_adapter=tools_adapter)
        self.azure_endpoint = None
        self.api_version = None
        self.api_key = None

    def _instantiate_openai_client(self, config):
        try:
            from openai import AzureOpenAI
            api_key_display = str(config.api_key)
            if api_key_display and len(api_key_display) > 8:
                api_key_display = api_key_display[:4] + "..." + api_key_display[-4:]
            client_kwargs = {
                "api_key": config.api_key,
                "azure_endpoint": getattr(config, "base_url", None),
                "api_version": config.extra.get("api_version", "2023-05-15"),
            }
            client = AzureOpenAI(**client_kwargs)
            return client
        except Exception as e:
            print(f"[ERROR] Exception during AzureOpenAI client instantiation: {e}", flush=True)
            import traceback
            print(traceback.format_exc(), flush=True)
            raise




from janito.tools.adapters.local import (
    local_tools_adapter as _internal_local_tools_adapter,
    LocalToolsAdapter,
)


def get_local_tools_adapter(workdir=None):
    # Use set_verbose_tools on the returned adapter to set verbosity as needed
    if workdir is not None:
        import os
        if not os.path.exists(workdir):
            os.makedirs(workdir, exist_ok=True)
        return LocalToolsAdapter(workdir=workdir)
    return _internal_local_tools_adapter


__all__ = [
    "LocalToolsAdapter",
    "get_local_tools_adapter",
]



from janito.llm.provider import LLMProvider
from janito.llm.model import LLMModelInfo
from janito.llm.auth import LLMAuthManager
from janito.llm.driver_config import LLMDriverConfig
from janito.tools.adapters.local.adapter import LocalToolsAdapter
from janito.providers.registry import LLMProviderRegistry

from .model_info import MODEL_SPECS

from janito.drivers.azure_openai.driver import AzureOpenAIModelDriver

available = AzureOpenAIModelDriver.available
unavailable_reason = AzureOpenAIModelDriver.unavailable_reason
maintainer = "João Pinto <lamego.pinto@gmail.com>"


class AzureOpenAIProvider(LLMProvider):
    name = "azure_openai"
    maintainer = "João Pinto <lamego.pinto@gmail.com>"
    MODEL_SPECS = MODEL_SPECS
    DEFAULT_MODEL = "azure_openai_deployment"

    def __init__(
        self, auth_manager: LLMAuthManager = None, config: LLMDriverConfig = None
    ):
        if not self.available:
            self._driver = None
            return
        self._auth_manager = auth_manager or LLMAuthManager()
        self._api_key = self._auth_manager.get_credentials(type(self).name)
        self._tools_adapter = LocalToolsAdapter()
        self._driver_config = config or LLMDriverConfig(model=None)
        if not self._driver_config.model:
            self._driver_config.model = self.DEFAULT_MODEL
        if not self._driver_config.api_key:
            self._driver_config.api_key = self._api_key
        if not self._driver_config.extra.get("api_version"):
            self._driver_config.extra["api_version"] = "2023-05-15"
        self.fill_missing_device_info(self._driver_config)
        self._driver = AzureOpenAIModelDriver(tools_adapter=self._tools_adapter)

    @property
    def driver(self):
        if not self.available:
            raise ImportError(
                f"AzureOpenAIProvider unavailable: {self.unavailable_reason}"
            )
        return self._driver

    @property
    def available(self):
        return available

    @property
    def unavailable_reason(self):
        return unavailable_reason

    def is_model_available(self, model_name):
        """
        Returns True for any model name, since Azure deployments are user-defined and not enumerable in advance.
        """
        return True

    def create_agent(self, tools_adapter=None, agent_name: str = None, **kwargs):
        from janito.llm.agent import LLMAgent
        from janito.drivers.azure_openai.driver import AzureOpenAIModelDriver

        # Always create a new driver with the passed-in tools_adapter
        driver = AzureOpenAIModelDriver(tools_adapter=tools_adapter)
        return LLMAgent(self, tools_adapter, agent_name=agent_name, **kwargs)

    def execute_tool(self, tool_name: str, event_bus, *args, **kwargs):
        # Use direct execution via adapter:
        self._tools_adapter.event_bus = event_bus
        return self._tools_adapter.execute_by_name(tool_name, *args, **kwargs)


LLMProviderRegistry.register(AzureOpenAIProvider.name, AzureOpenAIProvider)



from rich.console import Console
from rich.pretty import Pretty

from janito.config import config
from janito.cli.config import CONFIG_OPTIONS


def resolve_effective_model(provider_name):
    # Try provider-specific model, then global model, then provider default
    provider_cfg = config.get_provider_config(provider_name)
    model = provider_cfg.get("model") if provider_cfg else None
    if not model:
        model = config.get("model")
    if not model:
        try:
            from janito.provider_registry import ProviderRegistry

            provider_class = ProviderRegistry().get_provider(provider_name)
            model = getattr(provider_class, "DEFAULT_MODEL", None)
        except Exception:
            model = None
    return model


def handle_show_config(args):
    console = Console()
    provider = config.get("provider")
    model = config.get("model")
    # Show all providers with their effective model
    from janito.provider_registry import ProviderRegistry

    provider_names = []
    try:
        provider_names = ProviderRegistry()._get_provider_names()
    except Exception:
        pass
    console.print("[bold green]Current configuration:[/bold green]")
    console.print(f"[bold yellow]Current provider:[/bold yellow] {provider!r}\n")
    if model is not None:
        console.print(f"[bold yellow]Global model:[/bold yellow] {model!r}\n")
    if provider_names:
        console.print("[bold cyan]Provider specific default models:[/bold cyan]")
        for pname in provider_names:
            eff_model = resolve_effective_model(pname)
            prov_cfg = config.get_provider_config(pname)
            prov_model = prov_cfg.get("model") if prov_cfg else None
            extra = f" (override)" if prov_model else ""
            sel = "[default]" if pname == provider else ""
            console.print(
                f"  [bold]{pname}[/bold]{' '+sel if sel else ''}: model = [magenta]{eff_model}[/magenta]{extra}"
            )



from .base import ShellCmdHandler
from .edit import EditShellHandler
from .history_view import ViewShellHandler
from .lang import LangShellHandler
from .livelogs import LivelogsShellHandler
from .last import LastShellHandler
from .prompt import PromptShellHandler, RoleShellHandler, ProfileShellHandler
from .multi import MultiShellHandler
from .role import RoleCommandShellHandler
from .session import HistoryShellHandler
from .termweb_log import TermwebLogTailShellHandler
from .tools import ToolsShellHandler
from .help import HelpShellHandler
from janito.cli.console import shared_console

COMMAND_HANDLERS = {
    "/clear": __import__(
        "janito.cli.chat_mode.shell.commands.clear", fromlist=["ClearShellHandler"]
    ).ClearShellHandler,
    "/restart": __import__(
        "janito.cli.chat_mode.shell.commands.conversation_restart",
        fromlist=["RestartShellHandler"],
    ).RestartShellHandler,
    "/edit": EditShellHandler,
    "/view": ViewShellHandler,
    "/lang": LangShellHandler,
    "/livelogs": LivelogsShellHandler,
    "/last": LastShellHandler,
    "/prompt": PromptShellHandler,
    "/role": RoleShellHandler,
    "/profile": ProfileShellHandler,
    "/history": HistoryShellHandler,
    "/termweb-logs": TermwebLogTailShellHandler,
    "/tools": ToolsShellHandler,
    "/multi": MultiShellHandler,
    "/help": HelpShellHandler,
}


def get_shell_command_names():
    return sorted(cmd for cmd in COMMAND_HANDLERS.keys() if cmd.startswith("/"))


def handle_command(command, shell_state=None):
    parts = command.strip().split(maxsplit=1)
    cmd = parts[0]
    after_cmd_line = parts[1] if len(parts) > 1 else ""
    handler_cls = COMMAND_HANDLERS.get(cmd)
    if handler_cls:
        handler = handler_cls(after_cmd_line=after_cmd_line, shell_state=shell_state)
        return handler.run()
    shared_console.print(
        f"[bold red]Invalid command: {cmd}. Type /help for a list of commands.[/bold red]"
    )
    return None



from janito.llm.model import LLMModelInfo

MODEL_SPECS = {
    "azure_openai_deployment": LLMModelInfo(
        name="azure_openai_deployment",
        context="N/A",
        max_input="N/A",
        max_cot="N/A",
        max_response="N/A",
        thinking_supported=False,
        default_temp=0.2,
        open="azure_openai",
        driver="AzureOpenAIModelDriver",
    )
}



import time
import uuid
import traceback
import json
from typing import Optional, List, Dict, Any, Union
from janito.llm.driver import LLMDriver
from janito.driver_events import (
    GenerationStarted,
    GenerationFinished,
    RequestStarted,
    RequestFinished,
    ResponseReceived,
)
from janito.providers.openai.schema_generator import generate_tool_schemas
from janito.tools.adapters.local.adapter import LocalToolsAdapter
from janito.llm.message_parts import TextMessagePart, FunctionCallMessagePart
from janito.llm.driver_config import LLMDriverConfig

# Safe import of mistralai SDK
try:
    from mistralai import Mistral

    DRIVER_AVAILABLE = True
    DRIVER_UNAVAILABLE_REASON = None
except ImportError:
    DRIVER_AVAILABLE = False
    DRIVER_UNAVAILABLE_REASON = "Missing dependency: mistralai (pip install mistralai)"


class MistralAIModelDriver(LLMDriver):
    available = False
    unavailable_reason = "MistralAIModelDriver is not implemented yet."

    @classmethod
    def is_available(cls):
        return cls.available

    name = "mistralai"

    def __init__(self, tools_adapter=None):
        raise ImportError(self.unavailable_reason)



from janito.llm.model import LLMModelInfo

MODEL_SPECS = {
    "mistral-medium-latest": LLMModelInfo(
        name="mistral-medium-latest",
        context=32000,
        max_input=32000,
        max_cot="N/A",
        max_response=8192,
        thinking_supported=True,
        default_temp=0.2,
        open="mistralai",
        driver="MistralAIModelDriver",
    ),
    "mistral-large-latest": LLMModelInfo(
        name="mistral-large-latest",
        context=64000,
        max_input=64000,
        max_cot="N/A",
        max_response=16384,
        thinking_supported=True,
        default_temp=0.2,
        open="mistralai",
        driver="MistralAIModelDriver",
    ),
    "mistral-small-latest": LLMModelInfo(
        name="mistral-small-latest",
        context=16000,
        max_input=16000,
        max_cot="N/A",
        max_response=4096,
        thinking_supported=False,
        default_temp=0.2,
        open="mistralai",
        driver="MistralAIModelDriver",
    ),
}



import sys
import subprocess
import tempfile
import time
import http.client
import os
import threading
import queue
from rich.console import Console
from janito.i18n import tr
from janito.cli.config import get_termweb_port


def wait_for_termweb(port, timeout=3.0):
    """Polls the Bottle app root endpoint until it responds or timeout (seconds) is reached."""
    deadline = time.time() + timeout
    while time.time() < deadline:
        try:
            conn = http.client.HTTPConnection("localhost", port, timeout=0.5)
            conn.request("GET", "/")
            resp = conn.getresponse()
            if resp.status == 200:
                return True
        except Exception:
            pass
        time.sleep(0.1)
    return False


def termweb_start_and_watch(shell_state, shellstate_lock, selected_port=None):
    """
    Start the termweb server on the given port in a background thread.
    Communicates (termweb_proc, started: bool, stdout_path, stderr_path) via result_queue if provided.
    Returns the Thread object.
    """

    def termweb_worker(shell_state, shellstate_lock, selected_port, check_interval=2.0):
        """
        Worker to start termweb process, then monitor its running state/health, and update shell_state fields in a thread-safe manner.
        - shell_state: context or state object to update (must provide .termweb_pid, .termweb_status, etc.)
        - shellstate_lock: a threading.Lock or similar to synchronize access
        """
        console = Console()
        # Try to locate app.py
        app_py_path = os.path.join(os.path.dirname(__file__), "..", "termweb", "app.py")
        app_py_path = os.path.abspath(app_py_path)
        if not os.path.isfile(app_py_path):
            try:
                import janito_termweb

                app_py_path = janito_termweb.__file__.replace("__init__.py", "app.py")
            except ImportError:
                with shellstate_lock:
                    shell_state.termweb_status = "notfound"
                    shell_state.termweb_pid = None
                    shell_state.termweb_stdout_path = None
                    shell_state.termweb_stderr_path = None
                return
        termweb_stdout = tempfile.NamedTemporaryFile(
            delete=False, mode="w+", encoding="utf-8"
        )
        termweb_stderr = tempfile.NamedTemporaryFile(
            delete=False, mode="w+", encoding="utf-8"
        )
        port_to_use = selected_port if selected_port is not None else get_termweb_port()
        termweb_proc = subprocess.Popen(
            [sys.executable, app_py_path, "--port", str(port_to_use)],
            stdout=termweb_stdout,
            stderr=termweb_stderr,
        )

        # Step 1: Wait for server to become healthy (initial check)
        if wait_for_termweb(port_to_use, timeout=3.0):
            with shellstate_lock:
                shell_state.termweb_status = "running"
                shell_state.termweb_pid = termweb_proc.pid
                shell_state.termweb_stdout_path = termweb_stdout.name
                shell_state.termweb_stderr_path = termweb_stderr.name
                shell_state.termweb_port = port_to_use
        else:
            termweb_proc.terminate()
            termweb_proc.wait()
            # console.print(f"[red]Failed to start TermWeb on port {port_to_use}. Check logs for details.[/red]")
            with shellstate_lock:
                shell_state.termweb_status = "failed-start"
                shell_state.termweb_pid = None
                shell_state.termweb_stdout_path = termweb_stdout.name
                shell_state.termweb_stderr_path = termweb_stderr.name
                shell_state.termweb_port = None
            return

        # Step 2: Run watcher loop; exit and set fields if process or health fails
        import time
        from http.client import HTTPConnection

        while True:
            if termweb_proc.poll() is not None:  # means process exited
                with shellstate_lock:
                    shell_state.termweb_status = "terminated"
                    shell_state.termweb_pid = None
                break
            try:
                conn = HTTPConnection("localhost", port_to_use, timeout=0.5)
                conn.request("GET", "/")
                resp = conn.getresponse()
                if resp.status != 200:
                    raise Exception("Bad status")
            except Exception:
                # console.print(f"[red]TermWeb on port {port_to_use} appears to have stopped responding![/red]")
                with shellstate_lock:
                    shell_state.termweb_status = "unresponsive"
                break
            time.sleep(check_interval)

    # Launch background thread
    t = threading.Thread(
        target=termweb_worker,
        args=(shell_state, shellstate_lock, selected_port),
        daemon=True,
    )
    t.start()
    return t



from janito.cli.console import shared_console
from janito.cli.chat_mode.shell.commands.base import ShellCmdHandler


class LivelogsShellHandler(ShellCmdHandler):
    help_text = "Show the last lines of livereload logs. Usage: /livelogs [lines]"

    def run(self):
        lines_arg = self.after_cmd_line.strip()
        lines = 20
        if lines_arg and lines_arg.isdigit():
            lines = int(lines_arg)
        stdout_path = self.shell_state.termweb_stdout_path if self.shell_state else None
        stderr_path = (
            self.shell_state.livereload_stderr_path if self.shell_state else None
        )
        if not stdout_path and not stderr_path:
            shared_console.print(
                "[yellow][livereload] No livereload log files found for this session.[/yellow]"
            )
            return
        stdout_lines = []
        stderr_lines = []
        if stdout_path:
            try:
                with open(stdout_path, encoding="utf-8") as f:
                    stdout_lines = f.readlines()[-lines:]
                if stdout_lines:
                    shared_console.print(
                        f"[yellow][livereload][stdout] Tail of {stdout_path}:\n"
                        + "".join(stdout_lines)
                    )
            except Exception as e:
                shared_console.print(f"[red][livereload][stdout] Error: {e}[/red]")
        if stderr_path:
            try:
                with open(stderr_path, encoding="utf-8") as f:
                    stderr_lines = f.readlines()[-lines:]
                if stderr_lines:
                    shared_console.print(
                        f"[red][livereload][stderr] Tail of {stderr_path}:\n"
                        + "".join(stderr_lines)
                    )
            except Exception as e:
                shared_console.print(f"[red][livereload][stderr] Error: {e}[/red]")
        if (not stdout_path or not stdout_lines) and (
            not stderr_path or not stderr_lines
        ):
            shared_console.print("[livereload] No output or errors captured in logs.")



import subprocess
import os
import sys
import tempfile
import threading
from janito.tools.tool_base import ToolBase
from janito.report_events import ReportAction
from janito.tools.adapters.local.adapter import register_local_tool
from janito.i18n import tr


@register_local_tool
class PythonCodeRunTool(ToolBase):
    """
    Tool to execute Python code by passing it to the interpreter via standard input (stdin).

    Parameters:
        code (str): The Python code to execute as a string.
        timeout (int): Timeout in seconds for the command. Defaults to 60.

    Returns:
        str: Output and status message, or file paths/line counts if output is large.
    """
    provides_execution = True
    tool_name = "python_code_run"

    def run(self, code: str, timeout: int = 60) -> str:
        if not code.strip():
            self.report_warning(tr("ℹ️ Empty code provided."), ReportAction.EXECUTE)
            return tr("Warning: Empty code provided. Operation skipped.")
        self.report_action(
            tr("⚡ Running: python (stdin mode) ...\n{code}\n", code=code),
            ReportAction.EXECUTE,
        )
        self.report_stdout("\n")
        try:
            with (
                tempfile.NamedTemporaryFile(
                    mode="w+",
                    prefix="python_stdin_stdout_",
                    delete=False,
                    encoding="utf-8",
                ) as stdout_file,
                tempfile.NamedTemporaryFile(
                    mode="w+",
                    prefix="python_stdin_stderr_",
                    delete=False,
                    encoding="utf-8",
                ) as stderr_file,
            ):
                process = subprocess.Popen(
                    [sys.executable],
                    stdin=subprocess.PIPE,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                    text=True,
                    bufsize=1,
                    universal_newlines=True,
                    encoding="utf-8",
                    env={**os.environ, "PYTHONIOENCODING": "utf-8"},
                )
                stdout_lines, stderr_lines = self._stream_process_output(
                    process, stdout_file, stderr_file, code
                )
                return_code = self._wait_for_process(process, timeout)
                if return_code is None:
                    return tr(
                        "Code timed out after {timeout} seconds.", timeout=timeout
                    )
                stdout_file.flush()
                stderr_file.flush()
                self.report_success(
                    tr("✅ Return code {return_code}", return_code=return_code),
                    ReportAction.EXECUTE,
                )
                return self._format_result(
                    stdout_file.name, stderr_file.name, return_code
                )
        except Exception as e:
            self.report_error(tr("❌ Error: {error}", error=e), ReportAction.EXECUTE)
            return tr("Error running code via stdin: {error}", error=e)

    def _stream_process_output(self, process, stdout_file, stderr_file, code):
        stdout_lines = 0
        stderr_lines = 0

        def stream_output(stream, file_obj, report_func, count_func):
            nonlocal stdout_lines, stderr_lines
            for line in stream:
                file_obj.write(line)
                file_obj.flush()
                report_func(line.rstrip("\r\n"))
                if count_func == "stdout":
                    stdout_lines += 1
                else:
                    stderr_lines += 1

        stdout_thread = threading.Thread(
            target=stream_output,
            args=(process.stdout, stdout_file, self.report_stdout, "stdout"),
        )
        stderr_thread = threading.Thread(
            target=stream_output,
            args=(process.stderr, stderr_file, self.report_stderr, "stderr"),
        )
        stdout_thread.start()
        stderr_thread.start()
        process.stdin.write(code)
        process.stdin.close()
        stdout_thread.join()
        stderr_thread.join()
        return stdout_lines, stderr_lines

    def _wait_for_process(self, process, timeout):
        try:
            return process.wait(timeout=timeout)
        except subprocess.TimeoutExpired:
            process.kill()
            self.report_error(
                tr("❌ Timed out after {timeout} seconds.", timeout=timeout),
                ReportAction.EXECUTE,
            )
            return None

    def _format_result(self, stdout_file_name, stderr_file_name, return_code):
        with open(stdout_file_name, "r", encoding="utf-8", errors="replace") as out_f:
            stdout_content = out_f.read()
        with open(stderr_file_name, "r", encoding="utf-8", errors="replace") as err_f:
            stderr_content = err_f.read()
        max_lines = 100
        stdout_lines = stdout_content.count("\n")
        stderr_lines = stderr_content.count("\n")

        def head_tail(text, n=10):
            lines = text.splitlines()
            if len(lines) <= 2 * n:
                return "\n".join(lines)
            return "\n".join(
                lines[:n]
                + ["... ({} lines omitted) ...".format(len(lines) - 2 * n)]
                + lines[-n:]
            )

        if stdout_lines <= max_lines and stderr_lines <= max_lines:
            result = f"Return code: {return_code}\n--- python_code_run: STDOUT ---\n{stdout_content}"
            if stderr_content.strip():
                result += f"\n--- python_code_run: STDERR ---\n{stderr_content}"
            return result
        else:
            result = f"stdout_file: {stdout_file_name} (lines: {stdout_lines})\n"
            if stderr_lines > 0 and stderr_content.strip():
                result += f"stderr_file: {stderr_file_name} (lines: {stderr_lines})\n"
            result += f"returncode: {return_code}\n"
            result += (
                "--- python_code_run: STDOUT (head/tail) ---\n"
                + head_tail(stdout_content)
                + "\n"
            )
            if stderr_content.strip():
                result += (
                    "--- python_code_run: STDERR (head/tail) ---\n"
                    + head_tail(stderr_content)
                    + "\n"
                )
            result += "Use the view_file tool to inspect the contents of these files when needed."
            return result



from janito.llm.driver import LLMDriver
from janito.llm.driver_config import LLMDriverConfig
from janito.driver_events import (
    GenerationStarted,
    GenerationFinished,
    RequestStarted,
    RequestFinished,
    ResponseReceived,
)
from janito.llm.message_parts import TextMessagePart
import uuid
import traceback
import time

# Safe import of anthropic SDK
try:
    import anthropic

    DRIVER_AVAILABLE = True
    DRIVER_UNAVAILABLE_REASON = None
except ImportError:
    DRIVER_AVAILABLE = False
    DRIVER_UNAVAILABLE_REASON = "Missing dependency: anthropic (pip install anthropic)"


class AnthropicModelDriver(LLMDriver):
    available = False
    unavailable_reason = "AnthropicModelDriver is not implemented yet."

    @classmethod
    def is_available(cls):
        return cls.available

    """
    LLMDriver for Anthropic's Claude API (v3), using the anthropic SDK.
    """
    required_config = ["api_key", "model"]

    def __init__(self, tools_adapter=None):
        raise ImportError(self.unavailable_reason)

    def _create_client(self):
        try:
            import anthropic
        except ImportError:
            raise Exception(
                "The 'anthropic' Python SDK is required. Please install via `pip install anthropic`."
            )
        return anthropic.Anthropic(api_key=self.api_key)

    def _run_generation(
        self, messages_or_prompt, system_prompt=None, tools=None, **kwargs
    ):
        request_id = str(uuid.uuid4())
        client = self._create_client()
        try:
            prompt = ""
            if isinstance(messages_or_prompt, str):
                prompt = messages_or_prompt
            elif isinstance(messages_or_prompt, list):
                chat = []
                for msg in messages_or_prompt:
                    if msg.get("role") == "user":
                        chat.append("Human: " + msg.get("content", ""))
                    elif msg.get("role") == "assistant":
                        chat.append("Assistant: " + msg.get("content", ""))
                prompt = "\n".join(chat)
            if system_prompt:
                prompt = f"System: {system_prompt}\n{prompt}"

            self.publish(
                GenerationStarted,
                request_id,
                conversation_history=list(getattr(self, "_history", [])),
            )
            self.publish(RequestStarted, request_id, payload={})
            start_time = time.time()
            response = client.completions.create(
                model=self.model_name,
                max_tokens_to_sample=int(getattr(self.config, "max_response", 1024)),
                prompt=prompt,
                temperature=float(getattr(self.config, "default_temp", 0.7)),
            )
            duration = time.time() - start_time
            content = response.completion if hasattr(response, "completion") else None
            self.publish(
                RequestFinished,
                request_id,
                response=content,
                status=RequestStatus.SUCCESS,
                usage={},
            )
            parts = []
            if content:
                parts.append(TextMessagePart(content=content))
            self.publish(
                ResponseReceived,
                request_id=request_id,
                parts=parts,
                tool_results=[],
                timestamp=time.time(),
                metadata={"raw_response": response},
            )
            self.publish(GenerationFinished, request_id, total_turns=1)
        except Exception as e:
            self.publish(
                RequestFinished,
                request_id,
                status=RequestStatus.ERROR,
                error=str(e),
                exception=e,
                traceback=traceback.format_exc(),
            )



from janito.performance_collector import PerformanceCollector
from rich.tree import Tree
from rich.console import Console
from rich import print as rprint
import datetime

# TODO: Replace this with your actual collector instance retrieval
# For example, if you have a global or singleton:
# from janito.app_context import performance_collector as collector
from janito.perf_singleton import performance_collector as collector
from janito.cli.chat_mode.shell.commands.base import ShellCmdHandler
from janito.cli.console import shared_console


def _event_timestamp(event):
    if hasattr(event, "timestamp"):
        try:
            ts = float(getattr(event, "timestamp", 0))
            return f" [dim]{datetime.datetime.fromtimestamp(ts)}[/dim]"
        except Exception:
            return ""
    return ""


def _event_tool_name(event):
    return (
        f" [cyan]{getattr(event, 'tool_name', '')}[/cyan]"
        if hasattr(event, "tool_name")
        else ""
    )


def _event_params(event):
    return (
        f" Params: {getattr(event, 'params', '')}" if hasattr(event, "params") else ""
    )


def _event_result(event):
    return (
        f" Result: {getattr(event, 'result', '')}" if hasattr(event, "result") else ""
    )


def _event_error(event):
    return (
        f" [red]Error: {getattr(event, 'error')}[/red]"
        if hasattr(event, "error") and getattr(event, "error", None)
        else ""
    )


def _event_message(event):
    return (
        f" [yellow]Message: {getattr(event, 'message')}[/yellow]"
        if hasattr(event, "message")
        else ""
    )


def _event_subtype(event):
    return (
        f" [magenta]Subtype: {getattr(event, 'subtype')}[/magenta]"
        if hasattr(event, "subtype")
        else ""
    )


def _event_status(event):
    return (
        f" [blue]Status: {getattr(event, 'status')}[/blue]"
        if hasattr(event, "status")
        else ""
    )


def _event_duration(event):
    return (
        f" [green]Duration: {getattr(event, 'duration')}[/green]"
        if hasattr(event, "duration")
        else ""
    )


def format_event(event_tuple, parent_tree=None):
    event_type, event = event_tuple
    desc = f"[bold]{event_type}[/bold]"
    # Modular logic for each possible component
    desc += _event_timestamp(event)
    desc += _event_tool_name(event)
    desc += _event_params(event)
    desc += _event_result(event)
    desc += _event_error(event)
    desc += _event_message(event)
    desc += _event_subtype(event)
    desc += _event_status(event)
    desc += _event_duration(event)
    if parent_tree is not None:
        node = parent_tree.add(desc)
    else:
        node = Tree(desc)
    return node


def drill_down_last_generation():
    events = collector.get_all_events()
    # Find the last RequestStarted
    last_gen_start = None
    for i in range(len(events) - 1, -1, -1):
        if events[i][0] == "RequestStarted":
            last_gen_start = i
            break
    if last_gen_start is None:
        rprint("[red]No generations found.[/red]")
        return
    # Find the next GenerationFinished after last_gen_start
    for j in range(last_gen_start + 1, len(events)):
        if events[j][0] == "GenerationFinished":
            last_gen_end = j
            break
    else:
        last_gen_end = len(events) - 1
    gen_events = events[last_gen_start : last_gen_end + 1]
    tree = Tree("[bold green]Last Generation Details[/bold green]")
    for evt in gen_events:
        format_event(evt, tree)
    console = Console()
    console.print(tree)


class LastShellHandler(ShellCmdHandler):
    help_text = (
        "Show details of the last generation, with drill-down of tool executions."
    )

    def run(self):
        drill_down_last_generation()



from janito.llm.provider import LLMProvider
from janito.llm.model import LLMModelInfo
from janito.llm.auth import LLMAuthManager
from janito.llm.driver_config import LLMDriverConfig
from janito.drivers.google_genai.driver import GoogleGenaiModelDriver
from janito.tools.adapters.local.adapter import LocalToolsAdapter
from janito.providers.registry import LLMProviderRegistry

from .model_info import MODEL_SPECS

from janito.drivers.google_genai.driver import GoogleGenaiModelDriver

available = GoogleGenaiModelDriver.available
unavailable_reason = GoogleGenaiModelDriver.unavailable_reason
maintainer = "Needs maintainer"


class GoogleProvider(LLMProvider):
    MODEL_SPECS = MODEL_SPECS
    maintainer = "Needs maintainer"
    """
    Provider for Google LLMs via google-google.
    Default model: 'gemini-2.5-pro-preview-05-06'.
    """
    name = "google"
    DEFAULT_MODEL = "gemini-2.5-flash-preview-04-17"

    def __init__(self, config: LLMDriverConfig = None):
        if not self.available:
            self._driver = None
            return
        self._auth_manager = LLMAuthManager()
        self._api_key = self._auth_manager.get_credentials(type(self).name)
        self._tools_adapter = LocalToolsAdapter()
        self._info = config or LLMDriverConfig(model=None)
        if not self._info.model:
            self._info.model = self.DEFAULT_MODEL
        if not self._info.api_key:
            self._info.api_key = self._api_key
        self.fill_missing_device_info(self._info)
        self._driver = GoogleGenaiModelDriver(tools_adapter=self._tools_adapter)

    @property
    def driver(self) -> GoogleGenaiModelDriver:
        if not self.available:
            raise ImportError(f"GoogleProvider unavailable: {self.unavailable_reason}")
        return self._driver

    @property
    def available(self):
        return available

    @property
    def unavailable_reason(self):
        return unavailable_reason

    def create_agent(self, tools_adapter=None, agent_name: str = None, **kwargs):
        from janito.llm.agent import LLMAgent

        # Always create a new driver with the passed-in tools_adapter
        driver = GoogleGenaiModelDriver(tools_adapter=tools_adapter)
        return LLMAgent(self, tools_adapter, agent_name=agent_name, **kwargs)

    def execute_tool(self, tool_name: str, event_bus, *args, **kwargs):
        self._tools_adapter.event_bus = event_bus
        return self._tools_adapter.execute_by_name(tool_name, *args, **kwargs)


LLMProviderRegistry.register(GoogleProvider.name, GoogleProvider)



from janito.llm.model import LLMModelInfo

MODEL_SPECS = {
    "claude-3-opus-20240229": LLMModelInfo(
        name="claude-3-opus-20240229",
        max_response=200000,
        default_temp=0.7,
        driver="AnthropicModelDriver",
    ),
    "claude-3-sonnet-20240229": LLMModelInfo(
        name="claude-3-sonnet-20240229",
        max_response=200000,
        default_temp=0.7,
        driver="AnthropicModelDriver",
    ),
    "claude-3-haiku-20240307": LLMModelInfo(
        name="claude-3-haiku-20240307",
        max_response=200000,
        default_temp=0.7,
        driver="AnthropicModelDriver",
    ),
}



from janito.tools.tool_base import ToolBase
from janito.report_events import ReportAction


class SearchOutlineTool(ToolBase):
    """
    Tool for searching outlines in files.

    Args:
        file_path (str): Path to the file for which to generate an outline.
    Returns:
        str: Outline search result or status message.
    """

    tool_name = "search_outline"

    def run(self, file_path: str) -> str:
        from janito.tools.tool_utils import display_path
        from janito.i18n import tr

        self.report_action(
            tr(
                "🔍 Searching for outline in '{disp_path}'",
                disp_path=display_path(file_path),
            ),
            ReportAction.READ,
        )
        # ... rest of implementation ...
        # Example warnings and successes:
        # self.report_warning(tr("No files found with supported extensions."))
        # self.report_warning(tr("Error reading {file_path}: {error}", file_path=file_path, error=e))
        # self.report_success(tr("✅ {count} {match_word} found", count=len(output), match_word=pluralize('match', len(output))))
        pass



from typing import Type, Dict, Any
from janito.tools.tools_adapter import ToolsAdapterBase as ToolsAdapter


class LocalToolsAdapter(ToolsAdapter):
    def disable_execution_tools(self):
        """Unregister all tools with provides_execution = True."""
        to_remove = [name for name, entry in self._tools.items()
                     if getattr(entry["instance"], "provides_execution", False)]
        for name in to_remove:
            self.unregister_tool(name)

    """
    Adapter for local, statically registered tools in the agent/tools system.
    Handles registration, lookup, enabling/disabling, listing, and now, tool execution (merged from executor).
    """

    def __init__(self, tools=None, event_bus=None, allowed_tools=None, workdir=None):
        super().__init__(tools=tools, event_bus=event_bus, allowed_tools=allowed_tools)
        self._tools: Dict[str, Dict[str, Any]] = {}
        self.workdir = workdir
        if self.workdir:
            import os
            os.chdir(self.workdir)
        if tools:
            for tool in tools:
                self.register_tool(tool)

    def register_tool(self, tool_class: Type):
        instance = tool_class()
        if not hasattr(instance, "run") or not callable(instance.run):
            raise TypeError(
                f"Tool '{tool_class.__name__}' must implement a callable 'run' method."
            )
        tool_name = getattr(instance, "tool_name", None)
        if not tool_name or not isinstance(tool_name, str):
            raise ValueError(
                f"Tool '{tool_class.__name__}' must provide a class attribute 'tool_name' (str) for its registration name."
            )
        if tool_name in self._tools:
            raise ValueError(f"Tool '{tool_name}' is already registered.")
        self._tools[tool_name] = {
            "function": instance.run,
            "class": tool_class,
            "instance": instance,
        }

    def unregister_tool(self, name: str):
        if name in self._tools:
            del self._tools[name]

    def disable_tool(self, name: str):
        self.unregister_tool(name)

    def get_tool(self, name: str):
        return self._tools[name]["instance"] if name in self._tools else None

    def list_tools(self):
        return list(self._tools.keys())

    def get_tool_classes(self):
        return [entry["class"] for entry in self._tools.values()]

    def get_tools(self):
        return [entry["instance"] for entry in self._tools.values()]


    def add_tool(self, tool):
        # Register by instance (useful for hand-built objects)
        if not hasattr(tool, "run") or not callable(tool.run):
            raise TypeError(f"Tool '{tool}' must implement a callable 'run' method.")
        tool_name = getattr(tool, "tool_name", None)
        if not tool_name or not isinstance(tool_name, str):
            raise ValueError(
                f"Tool '{tool}' must provide a 'tool_name' (str) attribute."
            )
        if tool_name in self._tools:
            raise ValueError(f"Tool '{tool_name}' is already registered.")
        self._tools[tool_name] = {
            "function": tool.run,
            "class": tool.__class__,
            "instance": tool,
        }


# Optional: a local-tool decorator


def register_local_tool(tool=None):
    def decorator(cls):
        LocalToolsAdapter().register_tool(cls)
        return cls

    if tool is None:
        return decorator
    return decorator(tool)



# Provider static metadata registry for listing purposes (name, maintainer, and future fields)
STATIC_PROVIDER_METADATA = {
    "openai": {
        "maintainer": "João Pinto <lamego.pinto@gmail.com>",
    },
    "azure_openai": {
        "maintainer": "João Pinto <lamego.pinto@gmail.com>",
    },
    "google": {
        "maintainer": "Needs maintainer",
    },
    "mistralai": {
        "maintainer": "Needs maintainer",
    },
    "anthropic": {
        "maintainer": "Needs maintainer",
    },
    "deepseek": {
        "maintainer": "João Pinto <lamego.pinto@gmail.com>",
    },
}



from janito.tools.adapters.local.adapter import register_local_tool
from .python_outline import parse_python_outline
from .markdown_outline import parse_markdown_outline
from janito.formatting import OutlineFormatter
import os
from janito.tools.tool_base import ToolBase
from janito.report_events import ReportAction
from janito.tools.tool_utils import display_path, pluralize
from janito.i18n import tr

from janito.tools.adapters.local.adapter import register_local_tool as register_tool


@register_tool
class GetFileOutlineTool(ToolBase):
    """
    Get an outline of a file's structure. Supports Python and Markdown files.

    Args:
        file_path (str): Path to the file to outline.
    """

    tool_name = "get_file_outline"

    def run(self, file_path: str) -> str:
        try:
            self.report_action(
                tr(
                    "📄 Outline file '{disp_path}' ...",
                    disp_path=display_path(file_path),
                ),
                ReportAction.READ,
            )
            ext = os.path.splitext(file_path)[1].lower()
            with open(file_path, "r", encoding="utf-8", errors="replace") as f:
                lines = f.readlines()
            if ext == ".py":
                outline_items = parse_python_outline(lines)
                outline_type = "python"
                table = OutlineFormatter.format_outline_table(outline_items)
                self.report_success(
                    tr(
                        "✅ Outlined {count} {item_word}",
                        count=len(outline_items),
                        item_word=pluralize("item", len(outline_items)),
                    ),
                    ReportAction.READ,
                )
                return (
                    tr(
                        "Outline: {count} items ({outline_type})\n",
                        count=len(outline_items),
                        outline_type=outline_type,
                    )
                    + table
                )
            elif ext == ".md":
                outline_items = parse_markdown_outline(lines)
                outline_type = "markdown"
                table = OutlineFormatter.format_markdown_outline_table(outline_items)
                self.report_success(
                    tr(
                        "✅ Outlined {count} {item_word}",
                        count=len(outline_items),
                        item_word=pluralize("item", len(outline_items)),
                    ),
                    ReportAction.READ,
                )
                return (
                    tr(
                        "Outline: {count} items ({outline_type})\n",
                        count=len(outline_items),
                        outline_type=outline_type,
                    )
                    + table
                )
            else:
                outline_type = "default"
                self.report_success(
                    tr("✅ Outlined {count} items", count=len(lines)),
                    ReportAction.READ,
                )
                return tr(
                    "Outline: {count} lines ({outline_type})\nFile has {count} lines.",
                    count=len(lines),
                    outline_type=outline_type,
                )
        except Exception as e:
            self.report_error(
                tr("❌ Error reading file: {error}", error=e),
                ReportAction.READ,
            )
            return tr("Error reading file: {error}", error=e)

            if ext == ".py":
                outline_items = parse_python_outline(lines)
                outline_type = "python"
                table = OutlineFormatter.format_outline_table(outline_items)
                self.report_success(
                    tr(
                        "✅ Outlined {count} {item_word}",
                        count=len(outline_items),
                        item_word=pluralize("item", len(outline_items)),
                    ),
                    ReportAction.READ,
                )
                return (
                    tr(
                        "Outline: {count} items ({outline_type})\n",
                        count=len(outline_items),
                        outline_type=outline_type,
                    )
                    + table
                )
            elif ext == ".md":
                outline_items = parse_markdown_outline(lines)
                outline_type = "markdown"
                table = OutlineFormatter.format_markdown_outline_table(outline_items)
                self.report_success(
                    tr(
                        "✅ Outlined {count} {item_word}",
                        count=len(outline_items),
                        item_word=pluralize("item", len(outline_items)),
                    ),
                    ReportAction.READ,
                )
                return (
                    tr(
                        "Outline: {count} items ({outline_type})\n",
                        count=len(outline_items),
                        outline_type=outline_type,
                    )
                    + table
                )
            else:
                outline_type = "default"
                self.report_success(
                    tr("✅ Outlined {count} items", count=len(lines)),
                    ReportAction.READ,
                )
                return tr(
                    "Outline: {count} lines ({outline_type})\nFile has {count} lines.",
                    count=len(lines),
                    outline_type=outline_type,
                )
        except Exception as e:
            self.report_error(
                tr("❌ Error reading file: {error}", error=e),
                ReportAction.READ,
            )
            return tr("Error reading file: {error}", error=e)



import http.client
from rich.console import Console
from janito.cli.config import config
from janito.cli.console import shared_console
from janito.cli.chat_mode.shell.commands.base import ShellCmdHandler


class TermwebLogTailShellHandler(ShellCmdHandler):
    help_text = "Show the last lines of the latest termweb logs"

    def run(self):
        lines = 20
        args = self.after_cmd_line.strip().split()
        if args and args[0].isdigit():
            lines = int(args[0])
        stdout_path = self.shell_state.termweb_stdout_path if self.shell_state else None
        stderr_path = self.shell_state.termweb_stderr_path if self.shell_state else None
        status = getattr(self.shell_state, "termweb_status", None)
        live_status = getattr(self.shell_state, "termweb_live_status", None)
        status_checked = getattr(self.shell_state, "termweb_live_checked_time", None)
        shared_console.print(
            f"[bold cyan][termweb] Current run status: {status} | Last health check: {live_status} at {status_checked}"
        )
        if not stdout_path and not stderr_path:
            shared_console.print(
                "[yellow][termweb] No termweb log files found for this session.[/yellow]"
            )
            return
        stdout_lines = stderr_lines = None
        if stdout_path:
            try:
                with open(stdout_path, encoding="utf-8") as f:
                    stdout_lines = f.readlines()[-lines:]
                if stdout_lines:
                    shared_console.print(
                        f"[yellow][termweb][stdout] Tail of {stdout_path}:\n"
                        + "".join(stdout_lines)
                    )
            except Exception:
                pass
        if stderr_path:
            try:
                with open(stderr_path, encoding="utf-8") as f:
                    stderr_lines = f.readlines()[-lines:]
                if stderr_lines:
                    shared_console.print(
                        f"[red][termweb][stderr] Tail of {stderr_path}:\n"
                        + "".join(stderr_lines)
                    )
            except Exception:
                pass
        if (not stdout_path or not stdout_lines) and (
            not stderr_path or not stderr_lines
        ):
            shared_console.print("[termweb] No output or errors captured in logs.")


def handle_termweb_status(*args, shell_state=None, **kwargs):
    if shell_state is None:
        console.print(
            "[red]No shell state available. Cannot determine termweb status.[/red]"
        )
        return
    from janito.cli.config import get_termweb_port

    port = get_termweb_port()
    port_source = "config"
    pid = getattr(shell_state, "termweb_pid", None)
    stdout_path = getattr(shell_state, "termweb_stdout_path", None)
    stderr_path = getattr(shell_state, "termweb_stderr_path", None)
    running = False
    if port and hasattr(shell_state, "termweb_live_status"):
        running = shell_state.termweb_live_status == "online"
    console.print("[bold cyan]TermWeb Server Status:[/bold cyan]")
    console.print(f"  Running: {'[green]Yes[/green]' if running else '[red]No[/red]'}")
    if pid:
        console.print(f"  PID: {pid}")
    if port:
        console.print(f"  Port: {port} (from {port_source})")
        url = f"http://localhost:{port}/"
        console.print(f"  URL: [underline blue]{url}[/underline blue]")
    else:
        console.print("  [yellow]No port configured in config.[/yellow]")
    if stdout_path:
        console.print(f"  Stdout log: {stdout_path}")
    if stderr_path:
        console.print(f"  Stderr log: {stderr_path}")


handle_termweb_status.help_text = (
    "Show status information about the running termweb server"
)
